import numpy as np

from .base_optical_element import BaseOpticalElement
from .. import utils


class SiemensStar(BaseOpticalElement):
    """ A class representing a binary amplitude/phase Siemens Star.

    :param cycles: Number of Siemens Star cycles.
    :type cycles: int
    :param kind: Type of Siemens Star ["amplitude", "phase"].
    :type kind: str
    :param a_trans: Transmission of the Siemens Star [0-1].
    :type a_trans: float
    :param o_thick: Optical thickness of the Siemens Star.
    :type o_thick: float


    **Example**

    >>> import astropy.units as u
    >>> import pyoptica as po
    >>>
    >>> cycles = 5
    >>> star = po.SiemensStar(cycles)

    """

    VALID_RANGE_A_TRANS = (0, 1)
    VALID_RANGE_O_THICK = (0, 2 * np.pi)
    VALID_KIND = ["amplitude", "phase"]

    def __init__(self, cycles, kind='amplitude', a_trans=1, o_thick=np.pi):
        self.cycles = cycles
        self.kind = utils.param_validator(kind, self.VALID_KIND)
        self.a_trans = utils.range_validator(a_trans, self.VALID_RANGE_A_TRANS)
        self.o_thick = utils.range_validator(o_thick, self.VALID_RANGE_O_THICK)

    def amplitude_transmittance(self, wavefront):
        if "amplitude" in self.kind:
            _, phi = utils.cart2pol(wavefront.x, wavefront.y)
            return np.where(np.sin(self.cycles * phi) >= 0, 1, 0)
        else:
            return np.ones_like(wavefront.amplitude)

    def phase_transmittance(self, wavefront):
        if "phase" in self.kind:
            _, phi = utils.cart2pol(wavefront.x, wavefront.y)
            ss = np.where(np.sin(self.cycles * phi) >= 0, 1, 0)
            return np.exp(1.j * self.o_thick * ss)
        else:
            return np.exp(1.j * np.zeros_like(wavefront.phase))



import astropy.units as u
import numpy as np

from .base_optical_element import BaseOpticalElement


class CircularAperture(BaseOpticalElement):
    """ A class representing a circular diaphragm in an optical system.

    :param radius: radius of the aperture
    :type radius: astropy.Quantity of type length

    **Example**

    >>> import astropy.units as u
    >>> import pyoptica as po
    >>>
    >>> radius = 1 * u.mm
    >>> aperture = po.CircularAperture(radius)

    **References**

    [1] Joseph W. Goodman (2004) -
    *Introduction to Fourier Optics*, W. H. Freeman

    """

    @u.quantity_input(radius=u.m)
    def __init__(self, radius):
        self.radius = radius

    def amplitude_transmittance(self, wavefront):
        """Amplitude is fully transmitted for radii smaller than aperture radius.

        :param wavefront: The wavefront which interacts with the aperture
        :type wavefront: pyoptica.Wavefront
        :return: distribution of amplitude transmittance resulting from the aperture
        :rtype: numpy.array

        """
        x, y = wavefront.x, wavefront.y
        xy_squared = x ** 2 + y ** 2
        return np.where(xy_squared <= self.radius ** 2, 1, 0)

    def phase_transmittance(self, wavefront):
        r"""Phase is not disturbed for radii smaller than aperture radius.

        :param wavefront: The wavefront which interacts with the aperture
        :type wavefront: pyoptica.Wavefront
        :return: distribution of phase transmittance resulting from the aperture
        :rtype: numpy.array

        """

        return np.exp(1.j * np.zeros_like(wavefront.phase))


class CircularObscuration(CircularAperture):
    """ A class representing a circular obscuration in an optical system.

       :param radius: radius of the aperture
       :type radius: astropy.Quantity of type length

       **Example**

       >>> import astropy.units as u
       >>> import pyoptica as po
       >>>
       >>> radius = 1 * u.mm
       >>> obscuration = po.CircularObscuration(radius)

       **References**

       [1] Joseph W. Goodman (2004) -
       *Introduction to Fourier Optics*, W. H. Freeman

       """

    def amplitude_transmittance(self, wavefront):
        return 1.0 - super().amplitude_transmittance(wavefront)



import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm
from matplotlib.colors import LogNorm

from . import utils

FONT_SIZE = 8


class Plottable(object):
    """Base PyOptica class implementing plotting functionality.

    1. Implementing subclasses:
    Each subclass must implement `PLOTTING_OPTIONS` as a class constant as a
    dict with key names referring to class properties. Example:

    >>> import pyoptica as po
    >>>
    >>> class PlottableExample(po.plotting.Plottable):
    >>>     PLOTTING_OPTIONS = dict(
    >>>          first=dict(x='x', y='y', title=''), second=dict(x='x', y='y', title='')
    >>>    )
    >>>    @property
    >>>    def first(self):
    >>>         return np.ones((10, 10))
    >>>
    >>>    @property
    >>>     def second(self):
    >>>         return np.ones((10, 10))
    >>>
    >>>    @property
    >>>     def x(self):
    >>>         return np.meshgrid(np.arange(10), np.arange(10))[0] * u.m
    >>>
    >>>    @property
    >>>     def y(self):
    >>>         return np.meshgrid(np.arange(10), np.arange(10))[1] * u.m

    All plotables parameters (properties!) can be checked by running:

    >>> plottable_example = PlottableExample()
    >>> plottable_example.plottables
    ['first', 'second']

    The easiest way to plot them is to use the default plotting options (the
    ones defined in `PLOTTING_OPTIONS`):

    >>> plottable_example.plot(first='default')

    Or if we want to have a figure with 2 sublots:

    >>> plottable_example.plot(first='default', second='default')

    In case we would like to modify the default options it can be done by
    providing plotting options:

    >>> plottable_example.plot(first=dict(title='My Title', log_scale=True))

    Supported options are:

    axis_unit
        The X,Y grid will be converted and plotted in the unit. Expected astropy.Quantity of type length.
    bar_ticks
        Ticks of the colorbar
    bar_ticks_labels
        Labels of the ticks of the colorbar
    cmap
        matplotlib colormap to be used. Must be str!
    colorbar
        Should colorbar explaining values be shown (bool)
    colorbar_title
        Title of the colorbar
    x
        Name of the object property to be shown on X-axis
    y
        Name of the object property to be shown on Y-axis
    vmin
        Minimum value shown in the plot (Z-axis)
    vmax
        Maximum value shown in the plot (Z-axis)
    log_scale
        Log scale (bool)
    title
        Title of the plot
    z_function
        Function to be applied before plotting.

    Lastly, global properties of the plot can be modified with `fig_options`.
    All options accepted by matplotlib.figure are accepted.

    >>> plottable_example.plot(first='default', fig_options=dict(dpi=130))
    """
    FONT_SIZE = 8
    DEFAULT_PLOTTING_OPTIONS = dict(
        colorbar_title='[arb. u.]', bar_ticks=None, bar_ticks_labels=None,
        vmin=None, vmax=None, cmap='inferno', colorbar=True, z_function=None,
        log_scale=None
    )
    PLOTTING_OPTIONS = dict()

    @property
    def plottables(self):
        """ Returns a list of of all variable names that can be plotted
        :rtype: list of strings
        """
        return list(self.PLOTTING_OPTIONS.keys())

    @classmethod
    def _plot_img(
            cls, fig, ax, z, axis_unit, bar_ticks, bar_ticks_labels, cmap,
            colorbar, colorbar_title, fontsize, title, x1, x2, y1, y2, vmax,
            vmin, log_scale):
        """A base function for plotting a 2D distribution in matplotlib.

        :param fig: Figure used for plotting
        :type fig: matplotlib.pyplot.figure
        :param ax: Axis for the plot
        :type ax: matplotlib.pyplot.axis
        :param z: array to be plotted
        :type z: np.array
        :param axis_unit: The X,Y grid will be converted and plotted in the unit
        :type axis_unit: astropy.Quantity of type length
        :param bar_ticks: Ticks of the colorbar
        :type bar_ticks: iterable
        :param bar_ticks_labels: Labels of the ticks of the colorbar
        :type bar_ticks: iterable
        :param cmap: colormap to be used
        :type cmap: matplotlib.colors.LinearSegmentedColormap
        :param colorbar: Should colorbar explaining plotted values be shown
        :type colorbar: bool
        :param colorbar_title: Title of the colorbar
        :type colorbar_title: str
        :param fontsize: Size of the font
        :type fontsize: int
        :param title: Title of the plot
        :type title: str
        :param x1: minimum value on OX-axis
        :type x1: float
        :param x2: maximum value on OX-axis
        :type x2: float
        :param y1: minimum value on OY-axis
        :type y1: float
        :param y2: maximum value on OY-axis
        :type y2: float
        :param vmin: minimum value on OZ-axis
        :type vmin: float
        :param vmax: maximum value on OZ-axis
        :type vmax: float
        :param log_scale: should logarithmic scale be used>
        :type log_scale: bool
        :return: created image
        :rtype: matplotlib.image.AxesImage
        """
        cmap = cm.get_cmap(cmap)
        cmap.set_bad('k')
        axis_title = axis_unit.to_string()
        extent = [i.to(axis_unit).value for i in [x1, x2, y1, y2]]
        if log_scale is True:
            norm = LogNorm(vmin=vmin, vmax=vmax)
        else:
            norm = None
        im = ax.imshow(
            z, vmin=vmin, vmax=vmax, cmap=cmap, interpolation=None,
            extent=extent, norm=norm
        )
        ax.set_title(title)
        ax.set_xlabel(axis_title)
        ax.set_ylabel(axis_title)

        if colorbar is True:
            # The magic values make sure the colorbar looks good...
            # Found the on the internet.
            c_bar = fig.colorbar(im, fraction=0.046, pad=0.04, ax=ax)
            c_bar.ax.set_title(colorbar_title, fontsize=fontsize)
            if bar_ticks is not None:
                c_bar.set_ticks(bar_ticks)
            if bar_ticks_labels is not None:
                c_bar.set_ticklabels(bar_ticks_labels)
            c_bar.ax.tick_params(labelsize=fontsize)

        set_font_of_plot(ax, fontsize)
        return im

    def plot(self, fig_options=None, **kwargs):
        """ A method to plot 2D property of the object.

        :param fig_options: Global parameters referring to the entire plot
        :type fig_options: dict
        :param kwargs: a dict representing properties to be plotted with
                correnspodning options
        :return: fig, [axes], [images]

        **Example**

        >>> import astropy.units as u
        >>> import pyoptica as po
        >>>
        >>> wavelength = 500 * u.nm
        >>> pixel_scale = 22 * u.um
        >>> npix = 1024
        >>> wf = po.Wavefront(wavelength, pixel_scale, npix)
        >>> _ = wf.plot(intensity='default', phase='default')
        >>> _ = wf.plot(intensity=dict(title='My title', vmax=0.3, vmin=0.1))
        >>> _ = wf.plot(fig_options=dict(dpi=130), phase=dict(cmap='bwr'))
        """
        self._check_input(kwargs)

        to_plot = [k for k in kwargs.keys() if k in self.plottables]
        plots_no = len(to_plot)
        if plots_no < 1:
            raise ValueError(
                'Please choose at least one from: '
                f'{", ".join(self.PLOTTING_OPTIONS.keys())}.'
            )
        if fig_options is None:
            fig_options = dict()
        if 'font_size' in fig_options:
            font_size = fig_options.pop('font_size')
        else:
            font_size = FONT_SIZE
        fig, axs = plt.subplots(1, plots_no, **fig_options)
        if plots_no == 1:
            axs = [axs]
        ims = []

        for w, ax in zip(to_plot, axs):
            plotting_options = self.DEFAULT_PLOTTING_OPTIONS.copy()
            plotting_options.update(self.PLOTTING_OPTIONS[w])
            if kwargs[w] != 'default':
                plotting_options.update(kwargs[w])
            z = self.__getattribute__(w)
            z_function = plotting_options.pop('z_function')
            if z_function is not None:
                if isinstance(z_function, str):
                    z = self.__getattribute__(z_function)()
                elif callable(z_function):
                    z = z_function(z)
            x = self.__getattribute__(plotting_options.pop('x'))
            y = self.__getattribute__(plotting_options.pop('y'))
            plotting_options['x1'] = x[0, 0]
            plotting_options['x2'] = x[-1, -1]
            plotting_options['y1'] = y[0, 0]
            plotting_options['y2'] = y[-1, -1]
            plotting_options['fontsize'] = font_size
            plotting_options['axis_unit'] = plotting_options.get('axis_unit',
                                                                 x.unit)
            im = self._plot_img(fig, ax, z, **plotting_options)
            ims.append(im)
        plt.tight_layout()
        return fig, axs, ims

    def _check_plotting_options(self, options):
        """Checks if provided plotting options are supported

        :param options: provided plotting options
        :type options: dict
        :raises: ValueError
        """
        supported = dict(
            axis_unit='The X,Y grid will be converted and plotted in the unit.'
                      ' Expected astropy.Quantity.',
            bar_ticks='Ticks of the colorbar.',
            bar_ticks_labels='Labels of the ticks of the colorbar.',
            cmap='matplotlib colormap to be used. Must be str!',
            colorbar='Should colorbar explaining values be shown (bool).',
            colorbar_title='Title of the colorbar.',
            x='Name of the object property to be shown on X-axis.',
            y='Name of the object property to be shown on Y-axis.',
            vmin='Minimum value shown in the plot (Z-axis).',
            vmax='Maximum value shown in the plot (Z-axis).',
            log_scale='Log scale (bool).',
            title='Title of the plot.',
            z_function='Function to be applied before plotting.'
        )
        msg = '\n'.join([
            f'\t{k}\n\t\t{v}' for k, v in sorted(supported.items())
        ])
        for opt in options:
            if opt not in supported:
                raise ValueError(
                    f"Option '{opt}' is not supported. Please choose one of "
                    f"the following: \n{msg}"
                )

    def _check_input(self, input_kwargs):
        """Verifies if the input (parameters) can be plotted by the object.

        :param input_kwargs: parameters to be plotted
        :type input_kwargs: dict
        :raises: ValueError
        """
        if len(input_kwargs) == 0:
            raise ValueError(
                "At least one parameters from the following must be chosen: "
                f"{self.plottables}. \n\tExample: \n"
                f"\t\t object_instance.plot({self.plottables[0]}='default')\n"
                f"\tor\n"
                f"\t\t object_instance.plot({self.plottables[0]}=dict("
                f"log_scale=True, title='My Title'))"

            )
        for key in input_kwargs:
            if key not in self.plottables:
                raise ValueError(
                    f"{key} is not recognized as a plottable property of "
                    f"{self.__class__.__name__}. Only {self.plottables} "
                    f"are supported. The parameters can be checked with "
                    f"`obj_instance.plottables`")
            if input_kwargs[key] == 'default':
                continue
            self._check_plotting_options(input_kwargs[key])


def plot_zernike(
        zernike_array,
        rho,
        theta,
        axis_unit=u.dimensionless_unscaled,
        colorbar=True,
        zmin=None,
        zmax=None,
        bar_ticks=None,
        bar_ticks_labels=None,
        title=None,
        fontsize=FONT_SIZE,
        subplot_layout=111,
        dpi=130,
        figsize=(5, 5),
        cmap=None,
        log_scale=False,
        **kwargs
):
    """Plots Zernike distribution in matplotlib.

    :param zernike_array: array representing a polynomial
    :type zernike_array: numpy.array
    :param rho: radial coordinates of the plane,
    :type rho: numpy.array
    :param theta: angle coordinate of the plane
    :type theta: numpy.array
    :param axis_unit: The X,Y grid will be converted and plotted in the given unit
    :type axis_unit: astropy.Quantity of type length
    :param colorbar: Should colorbar explaining plotted values be shown
    :type colorbar: bool
    :param zmin: Minimal value to be shown
    :type zmin: float
    :param zmax: Maximal value to be shown
    :type zmax: float
    :param bar_ticks: Ticks of the colorbar
    :type bar_ticks: iterable
    :param bar_ticks_labels: Labels of the ticks of the colorbar
    :type bar_ticks: iterable
    :param title: Title of the plot
    :type title: str
    :param fontsize: Size of the font
    :type fontsize: int
    :param figsize: Size of output figure in inches
    :type figsize: (int, int)
    :param dpi: dots per inch of the resulting figure
    :type dpi: int
    :param cmap: colormap to be used
    :type cmap: matplotlib.colors.LinearSegmentedColormap
    :param log_scale: should logarithmic scale be used?
    :type: bool
    :param kwargs: all kwargs passed to matplotlib.pyplot.imshow
    :type kwargs: object
    :return: plot of asked values
    :rtype: matplotlib.figure, matplotlib.axis

    **Example**

    >>> import astropy.units as u
    >>> import numpy as np
    >>> import pyoptica as po
    >>>
    >>> npix = 101
    >>> pixel_scale = 0.5 * u.m
    >>> x, y = po.utils.mesh_grid(npix, pixel_scale)
    >>> r_max = .48 * npix * pixel_scale
    >>> r, theta = po.utils.cart2pol(x, y)
    >>> r = r / r_max
    >>>
    >>> m, n = -1, 3
    >>> z = po.zernike((m, n), 'mn', r, theta, fill_value=np.nan)
    >>> _ = po.plotting.plot_zernike(z, r, theta, title=f'Zernike m={m}, n={n}')

    """
    if title is None:
        title = 'Zernike Distribution'
    x, y = utils.pol2cart(rho, theta)
    x1, x2 = x[0, 0], x[-1, -1]
    y1, y2 = y[0, 0], y[-1, -1]
    if cmap is None:
        cmap = cm.plasma
    if zmin is None and zmax is None:
        zmax = np.nanmax(zernike_array)
        zmin = np.nanmin(zernike_array)
        zmax = max(abs(zmax), abs(zmin))
        zmin = -zmax
    colorbar_title = '[arb. u.]'
    return plot_img(
        zernike_array, axis_unit, bar_ticks, bar_ticks_labels, cmap, colorbar,
        colorbar_title, dpi, figsize, fontsize, subplot_layout, title, x1, x2,
        y1, y2, zmax, zmin, log_scale, **kwargs)


def plot_img(
        z, axis_unit, bar_ticks, bar_ticks_labels, cmap, colorbar,
        colorbar_title, dpi, figsize, fontsize, subplot_layout, title, x1, x2,
        y1, y2, zmax, zmin, log_scale, **kwargs):
    """A base function for plotting a 2D distribution in matplotlib.
    :param z: array to be plotted
    :type z: np.array
    :param axis_unit: The X,Y grid will be converted and plotted in the unit
    :type axis_unit: astropy.Quantity of type length
    :param bar_ticks: Ticks of the colorbar
    :type bar_ticks: iterable
    :param bar_ticks_labels: Labels of the ticks of the colorbar
    :type bar_ticks: iterable
    :param cmap: colormap to be used
    :type cmap: matplotlib.colors.LinearSegmentedColormap
    :param colorbar: Should colorbar explaining plotted values be shown
    :type colorbar: bool
    :param colorbar_title: Title of the colorbar
    :type colorbar_title: str
    :param dpi: dots per inch of the resulting figure
    :type dpi: int
    :param figsize: Size of output figure in inches
    :type figsize: (int, int)
    :param fontsize: Size of the font
    :type fontsize: int
    :param subplot_layout: layout in case of many plots
    :type subplot_layout: int
    :param title: Title of the plot
    :type title: str
    :param x1: minimum value on OX-axis
    :type x1: float
    :param x2: maximum value on OX-axis
    :type x2: float
    :param y1: minimum value on OY-axis
    :type y1: float
    :param y2: maximum value on OY-axis
    :type y2: float
    :param zmin: minimum value on OZ-axis
    :type zmin: float
    :param zmax: maximum value on OZ-axis
    :type zmax: float
    :param log_scale: should logarithmic scale be used>
    :type log_scale: bool
    :param kwargs: all kwargs passed to matplotlib.pyplot.imshow
    :type kwargs: object
    :return: plot of asked values
    :rtype: matplotlib.figure, matplotlib.axis

    """
    axis_title = axis_unit.to_string()
    fig = plt.figure(dpi=dpi, figsize=figsize)
    ax = fig.add_subplot(subplot_layout)
    extent = [i.to(axis_unit).value for i in [x1, x2, y1, y2]]
    if log_scale is True:
        norm = LogNorm(vmin=zmin, vmax=zmax)
    else:
        norm = None
    plot = ax.imshow(
        z, vmin=zmin, vmax=zmax, cmap=cmap, interpolation=None, extent=extent,
        norm=norm, **kwargs
    )
    ax.set_title(title)
    ax.set_xlabel(axis_title)
    ax.set_ylabel(axis_title)
    if colorbar is True:
        # The magic values make sure the colorbar looks good...
        # Found the on the internet.
        c_bar = plt.colorbar(plot, fraction=0.046, pad=0.04)
        c_bar.ax.set_title(colorbar_title, fontsize=fontsize)
        if bar_ticks is not None:
            c_bar.set_ticks(bar_ticks)
        if bar_ticks_labels is not None:
            c_bar.set_ticklabels(bar_ticks_labels)
        c_bar.ax.tick_params(labelsize=fontsize)

    set_font_of_plot(ax, fontsize)
    plt.tight_layout()
    return fig, plot, ax


def set_font_of_plot(plot, font_size=FONT_SIZE):
    """Sets font of a plot
    :param plot: matplotlib.axes
        Plot on which font should be adjusted
    :param font_size: int

    """
    plot.title.set_fontsize(font_size)
    labels = [plot.xaxis.label, plot.yaxis.label] + \
             plot.get_xticklabels() + plot.get_yticklabels()
    for item in labels:
        item.set_fontsize(font_size)



import logging

LOGGING_LEVEL = logging.INFO
LOGGING_FORMAT = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'


class LoggerMixin(object):
    """A mixin class to add logger. It is expected to be used in all classes
    developed withing pyoptica. It is expected to be used only internally!

    **Example**

    >>> import pyoptica.logging as pol
    >>>
    >>> class MyClass(pol.LoggerMixin):
    >>>     pass

    >>> my_instance = MyClass()
    >>> my_instance.logger.info('I am printing a log.')
    yyy-mm-dd hh:mm:ss,sss - module_name.MyClass - INFO - I am printing a log.
    """

    def __init_subclass__(cls, **kwargs):
        name = '.'.join([cls.__module__, cls.__name__])
        logger = get_standard_logger(name)
        cls.logger = logger


def get_standard_logger(name, level=None):
    """ Gets the logger with provided name` with default values for the project

    :param name: name of the logger
    :type name: str
    :param level: level of logging (default logging.DEBUG)
    :type level: int
    :return: logger
    :rtype: logging.Logger

    **Example**

    >>> import pyoptica.logging as pol
    >>>
    >>> pol.get_standard_logger('my_name')
    <Logger my_name (DEBUG)>
    """
    level = level or LOGGING_LEVEL
    logger = logging.getLogger(name)
    logger.setLevel(level)
    ch = logging.StreamHandler()
    ch.setLevel(LOGGING_LEVEL)
    formatter = logging.Formatter(LOGGING_FORMAT)
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger



from . import distribution
from . import holography
from . import image
from . import imaging_system
from . import logging
from . import optical_elements
from . import plotting
from . import utils
from . import wavefront
from . import zernikes
from .image import Image
from .imaging_system import *
from .optical_elements import *
from .wavefront import *
from .zernikes import *

__all__ = [
    'logging',
    'distribution',
    'wavefront',
    'Wavefront',
    'optical_elements',
    'plotting',
    'utils',
    'imaging_system',
    'ImagingSystem',
    'zernikes',
    'holography',
]
__all__ += optical_elements.__all__
__all__ += zernikes.__all__

__version__ = '0.6.2'



import astropy.units as u
import numpy as np
from astropy.io import fits

from . import distribution


class Wavefront(distribution.Distribution):
    """A class representing a wavefront.

    :param wavelength: Wavelength of simulated wavefront
    :type wavelength: astropy.Quantity of type length
    :param pixel_scale: Sampling of the wavefront
    :type pixel_scale: astropy.Quantity of type length
    :param npix: Number of pixels in the wavefront array
    :type npix: int

    :raises TypeError: wavelength and pixelscale are expected to be of type
        astropy.quantity of type length

    **Example**

    >>> import astropy.units as u
    >>> import pyoptica as po
    >>>
    >>> wavelength = 500 * u.nm
    >>> pixel_scale = 22 * u.um
    >>> npix = 1024
    >>> wf = po.Wavefront(wavelength, pixel_scale, npix)

    **References**

    [1] Joseph W. Goodman (2004) - *Introduction to Fourier Optics*, W. H. Freeman

    [2] Kedar Khare (2016) - *Fourier Optics and Computational Imaging*, Wiley&Sons Ltd.

    [3] David Voelz (2011) - *Computational Fourier Optics Matlab Tutorial*, Spie Press

    """
    PLOTTING_OPTIONS = {
        'intensity': dict(x='x', y='y', title='Intensity'),
        'amplitude': dict(x='x', y='y', title='Amplitude'),
        'phase': dict(
            colorbar_title='[rad]', vmin=-np.pi, vmax=np.pi,
            bar_ticks=[-np.pi, -np.pi / 2, 0, np.pi / 2, np.pi],
            bar_ticks_labels=[
                r'$-\pi$',
                r'$-\frac{\pi}{2}$',
                r'$0$',
                r'$\frac{\pi}{2}$',
                r'$\pi$'
            ],
            x='x', y='y', cmap='hsv', title='Phase',
            z_function='_apply_phase_plotting_threshold'
        )
    }
    @u.quantity_input(wavelength=u.m, pixel_scale=u.m)
    def __init__(self, wavelength, pixel_scale, npix):
        super().__init__(wavelength, pixel_scale, npix)
        self.wavefront = np.ones((npix, npix), dtype=np.complex64)
        self.logger.debug(f"Created {self}")

    @property
    def amplitude(self):
        """Get/Set amplitude of the wavefront. Taking/returning
        an array of shape=(npix, npix).

        :getter: Amplitude of the wavefront.
        :setter: Loads given array as amplitude of the wavefront,
            leaving phase unchanged.
        :type: np.array
        """
        return np.abs(self.wavefront)

    @amplitude.setter
    def amplitude(self, amplitude):
        self.wavefront = amplitude * np.exp(1.j * self.phase)

    @property
    def intensity(self):
        """Get/Set Intensity of the wavefront. Taking/returning
        an array of shape=(npix, npix) .

        :getter: Intensity of the wavefront.
        :setter: Loads given array as intensity of the wavefront,
            leaving phase unchanged.
        :type: np.array

        """
        return self.amplitude ** 2

    @intensity.setter
    def intensity(self, intensity):
        self.wavefront = np.sqrt(intensity) * np.exp(1.j * self.phase)

    @property
    def phase(self):
        """Get/Set Phase of the wavefront. Taking/returning
        an array of shape=(npix, npix).

        :getter: Phase of the wavefront.
        :setter: Loads given array as phase of the wavefront,
            leaving amplitude unchanged.

        :type: np.array

        """
        return np.angle(self.wavefront)

    @phase.setter
    def phase(self, phase):
        self.wavefront = self.amplitude * np.exp(1.j * phase)

    def to_fits(self, path):
        """A method to serialize wavefront to fits. Amplitude and phase are saved
        separately to seperate tables

        :param path: path to the file to which wavefront is saved
        :type path: str

        **Example**

        >>> import astropy.units as u
        >>> import pyoptica as po
        >>>
        >>> wavelength = 500 * u.nm
        >>> pixel_scale = 22 * u.um
        >>> npix = 1024
        >>> wf = po.Wavefront(wavelength, pixel_scale, npix)
        >>> file_path = 'path/to/your/file.fits'
        >>> wf.to_fits(file_path)

        """
        amplitude = self.amplitude
        phase = self.phase
        hdu_list = fits.HDUList()
        hdr = fits.Header()
        hdr['WVL'] = self.wavelength.to(u.nm).value,
        hdr['PIXSCALE'] = self.pixel_scale.to(u.um).value,
        hdr['NPIX'] = self.npix
        hdu_list.append(fits.PrimaryHDU(header=hdr))
        hdu_list.append(fits.ImageHDU(amplitude, header=None, name='AMP',
                                      do_not_scale_image_data=True))
        hdu_list.append(fits.ImageHDU(phase, header=None, name='PHS',
                                      do_not_scale_image_data=True))
        hdu_list.writeto(path, overwrite=True)
        self.logger.debug(f"Saved wavefront to {path}")

    @classmethod
    def from_fits(cls, path):
        """A method to load wavefront from a fits file.

        :param path: path to the file to be loaded.
        :type path: str
        :return: loaded wavefront
        :rtype: pyoptica.Wavefront

        **Example**

        >>> import pyoptica as po
        >>>
        >>> file_path = 'path/to/your/file.fits'
        >>> wf = po.Wavefront.from_fits(file_path)

        """
        hdu_list = fits.open(path, memmap=True)
        wavelength = hdu_list[0].header['WVL'] * u.nm
        pixel_scale = hdu_list[0].header['PIXSCALE'] * u.um
        npix = hdu_list[0].header['NPIX']
        wf = Wavefront(wavelength, pixel_scale, npix)
        wf.amplitude = hdu_list['AMP'].data
        wf.phase = hdu_list['PHS'].data
        cls.logger.debug(f"Loaded wavefront from {path}")
        return wf

    def _apply_phase_plotting_threshold(self):
        # Phase is plotted only for non-zero intensity values
        avg_i = np.average(self.intensity[self.intensity > 0])
        no_phase_indices = np.where(
            self.intensity < avg_i * 0.01
        )
        phase = self.phase
        phase[no_phase_indices] = np.nan
        return phase



import astropy.units as u
import numpy as np
from astropy.io import fits

from . import distribution


class Image(distribution.Distribution):
    """Class representing an Image -- result of the imaging operation.
    In addition to the image it stores information regarding wavelength,
    pixel sampling, and dimensions of the image distribution.

    :param wavelength: Wavelength of the simulated image
    :type wavelength: astropy.Quantity of type length
    :param pixel_scale: Sampling of the image
    :type pixel_scale: astropy.Quantity of type length
    :param npix: Number of pixels in the image array
    :type npix: int

    :raises TypeError: wavelength and pixelscale are expected to be of type
        astropy.quantity of type length

    **Example**

    >>> import astropy.units as u
    >>> import pyoptica as po
    >>>
    >>> wavelength = 500 * u.nm
    >>> pixel_scale = 22 * u.um
    >>> npix = 1024
    >>> image = po.Image(wavelength, pixel_scale, npix)

    """
    PLOTTING_OPTIONS = {
        'image': dict(x='x', y='y', title='Image')
    }
    @u.quantity_input(wavelength=u.m, pixel_scale=u.m)
    def __init__(self, wavelength, pixel_scale, npix):
        super().__init__(wavelength, pixel_scale, npix)
        self.image = np.ones((npix, npix), dtype=np.float)
        self.logger.debug(f"Created {self}")

    def to_fits(self, path):
        """A method to serialize image to fits.

        :param path: path to the file to which image is saved
        :type path: str

        **Example**

        >>> import astropy.units as u
        >>> import pyoptica as po
        >>>
        >>> wavelength = 500 * u.nm
        >>> pixel_scale = 22 * u.um
        >>> npix = 1024
        >>> image = po.Image(wavelength, pixel_scale, npix)
        >>> file_path = 'path/to/your/file.fits'
        >>> image.to_fits(file_path)

        """
        image = self.image
        hdu_list = fits.HDUList()
        hdr = fits.Header()
        hdr['WVL'] = self.wavelength.to(u.nm).value,
        hdr['PIXSCALE'] = self.pixel_scale.to(u.um).value,
        hdr['NPIX'] = self.npix
        hdu_list.append(fits.PrimaryHDU(header=hdr))
        hdu_list.append(fits.ImageHDU(image, header=None, name='IMG',
                                      do_not_scale_image_data=True))
        hdu_list.writeto(path, overwrite=True)
        self.logger.debug(f"Saved wavefront to {path}")

    @classmethod
    def from_fits(cls, path):
        """A method to load image from a fits file.

        :param path: path to the file to be loaded.
        :type path: str
        :return: loaded image
        :rtype: pyoptica.Image

        **Example**

        >>> import pyoptica as po
        >>>
        >>> file_path = 'path/to/your/file.fits'
        >>> image = po.Image.from_fits(file_path)

        """
        hdu_list = fits.open(path, memmap=True)
        wavelength = hdu_list[0].header['WVL'] * u.nm
        pixel_scale = hdu_list[0].header['PIXSCALE'] * u.um
        npix = hdu_list[0].header['NPIX']
        image = Image(wavelength, pixel_scale, npix)
        image.image = hdu_list['IMG'].data
        cls.logger.debug(f"Loaded image from {path}")
        return image




import astropy.units as u
import numpy as np

from .base_optical_element import BaseOpticalElement


class RectangularAperture(BaseOpticalElement):
    """ A class representing a rectangular aperture in an optical system.

    :param width: width of the aperture
    :type width: astropy.Quantity of type length
    :param height: heigth of the aperture
    :type height: astropy.Quantity of type length

    **Example**

    >>> import astropy.units as u
    >>> import pyoptica as po
    >>>
    >>> width = 1 * u.mm
    >>> height = 1 * u.mm
    >>> aperture = po.RectangularAperture(width, height)

    **References**

    [1] Joseph W. Goodman (2004) -
    *Introduction to Fourier Optics*, W. H. Freeman

    """

    @u.quantity_input(width=u.m, height=u.m)
    def __init__(self, width, height):
        self.width = width
        self.height = height

    def amplitude_transmittance(self, wavefront):
        """Amplitude is fully transmitted for dimensions smaller than aperture size.

        :param wavefront: The wavefront which interacts with the aperture
        :type wavefront: pyoptica.Wavefront
        :return: distribution of amplitude transmittance resulting from the aperture
        :rtype: numpy.array

        """
        x, y = wavefront.x, wavefront.y
        return np.where((np.abs(x) <= (.5 * self.width)), 1, 0) * \
               np.where(np.abs(y) <= (.5 * self.height), 1, 0)

    def phase_transmittance(self, wavefront):
        r"""Phase is not disturbed for dimensions smaller than aperture radius.

        :param wavefront: The wavefront which interacts with the aperture
        :type wavefront: pyoptica.Wavefront
        :return: distribution of phase transmittance resulting from the aperture
        :rtype: numpy.array

        """

        return np.exp(1.j * np.zeros_like(wavefront.phase))


class RectangularObscuration(RectangularAperture):
    """ A class representing a rectangular obscuration in an optical system.

    :param width: width of the aperture
    :type width: astropy.Quantity of type length
    :param height: heigth of the aperture
    :type height: astropy.Quantity of type length

    **Example**

    >>> import astropy.units as u
    >>> import pyoptica as po
    >>>
    >>> width = 1 * u.mm
    >>> height = 1 * u.mm
    >>> obscuration = po.RectangularObscuration(width, height)

    **References**

    [1] Joseph W. Goodman (2004) -
    *Introduction to Fourier Optics*, W. H. Freeman

    """

    def amplitude_transmittance(self, wavefront):
        return 1.0 - super().amplitude_transmittance(wavefront)



import warnings

import astropy.units as u
import numpy as np

from .base_optical_element import BaseOpticalElement


class ThinLens(BaseOpticalElement):
    """ A class representing a thin lens. The reasoning used to develop
    functionality of the lens is based on chapter 5 of [1].

    :param radius: radius of the lens
    :type radius: astropy.Quantity of type length
    :param f: focal distance of the lens
    :type f: astropy.Quantity of type length

    **Example**

    >>> import astropy.units as u
    >>> import pyoptica as po
    >>>
    >>> f = 1 * u.cm
    >>> radius = 1 * u.mm
    >>> lens = po.ThinLens(radius, f)

    **References**

    [1] Joseph W. Goodman (2004) -
    *Introduction to Fourier Optics*, W. H. Freeman

    """

    @u.quantity_input(radius=u.m, f=u.m)
    def __init__(self, radius, f):
        self.radius = radius
        self.f = f

    def amplitude_transmittance(self, wavefront):
        """In the thin lens approximation amplitude is fully transmitted.

        :param wavefront: The wavefront which interacts with the lens
        :type wavefront: pyoptica.Wavefront
        :return: distribution of amplitude transmittance resulting from the lens
        :rtype: numpy.array

        """
        return np.ones_like(wavefront.amplitude)

    def phase_transmittance(self, wavefront):
        r"""Calculations of phase transmittance based on eq. 5.10 in [1]:

        :math:`t_{l}(x, y)=\exp \left[-j \frac{k}{2 f}\left(x^{2}+y^{2}\right)\right]`

        :param wavefront: The wavefront which interacts with the lens
        :type wavefront: pyoptica.Wavefront
        :return: distribution of phase transmittance resulting from the lens
        :rtype: numpy.array

        """
        self._check_sampling(wavefront)
        k = wavefront.k
        x, y = wavefront.x, wavefront.y
        xy_squared = x ** 2 + y ** 2
        t1 = np.exp(-(1.j * k) / (2 * self.f) * xy_squared)
        phi = np.where(
            xy_squared <= self.radius ** 2, t1, 1
        )
        return phi

    def _check_sampling(self, wavefront):
        """Checks if sampling of the given wavefront meets the requirement:

        :matha: `\Delta x  \leq \frac{|f|}{D_{L}} \lambda`

        Based on eq. 6.14 in  David Voelz (2011)
        *Computational Fourier Optics Matlab Tutorial*, Spie Press

        :param wavefront: wavefront to be multiplied by the lens
        :type wavefront: pyoptica.Wavefront

        """
        min_sampling = np.abs(self.f) / (
                    2 * self.radius) * wavefront.wavelength
        # Just for readable output.
        min_sampling = min_sampling.to(wavefront.pixel_scale.unit)
        if wavefront.pixel_scale > min_sampling:
            warning = f'Bad sampling: {wavefront.pixel_scale} > {min_sampling}'
            self.logger.warning(warning)
            warnings.warn(warning)



from math import ceil, factorial, floor, sqrt

import astropy.units as u
import numpy as np
from scipy.optimize import minimize

from . import logging

__all__ = [
    'zernike', 'fit_zernikes', 'fit_zernikes_lstsq', 'construct_wavefront',
    'get_zernike_name', 'mn_to_noll', 'mn_to_fringe', 'mn_to_osa',
    'noll_to_mn', 'osa_to_mn', 'fringe_to_mn'
]

logger = logging.get_standard_logger(__name__)

ZERNIKE_NAMES = {
    # Based on *Wavefront Optics for Vision Correction* by Guang-ming Dai
    (0, 0): 'Piston', (-1, 1): 'y-Tilt', (1, 1): 'x-Tilt',
    (-2, 2): 'y-Astigmatism', (0, 2): 'Defocus', (2, 2): 'x-Astigmatism',
    (-3, 3): 'y-Trefoil', (-1, 3): 'y-Coma', (1, 3): 'x-Coma',
    (3, 3): 'x-Trefoil', (-4, 4): 'y-Quadrafoil',
    (-2, 4): 'y-Secondary Astigmatism', (0, 4): 'Spherical Aberration',
    (2, 4): 'x-Secondary Astigmatism', (4, 4): 'x-Quadrafoil'
}  # I use mn to avoid any confusion about indexing!


def zernike(index, convention, rho, theta, normalize=True, fill_value=0):
    """Calculates zernike polynomial for given index (or indices) for
    convention (`mn`, `noll`, `osa`). Implementation based on [1]. If
     normalize is `True` output is normalized following Noll's convention [2].

    :param index: index (or indices for mn) of the polynomial
    :type index: int or tuple of ints
    :param convention: used indexing convention
    :type convention: str
    :param rho: radial coordinates of the plane,
    :type rho: numpy.array
    :param theta: angle coordinate of the plane
    :type theta: numpy.array
    :param normalize: Should output be normalized following Noll's convention?
    :type normalize: bool
    :param fill_value: what value should be put for r > 1
    :type fill_value: numeric
    :return: resulting _zernike_mn distribution
    :rtype: numpy.array


    **Example**

    >>> import astropy.units as u
    >>> import pyoptica as po
    >>>
    >>> npix = 101
    >>> pixel_scale = 0.5 * u.m
    >>> x, y = po.utils.mesh_grid(npix, pixel_scale)
    >>> r_max = .5 * npix * pixel_scale
    >>> r, theta = po.utils.cart2pol(x, y)
    >>> r = r / r_max
    >>> j = 8
    >>> z_8 = po.zernikes.zernike(j, 'noll', r, theta)

    In case we want to try different convention:

    >>> z_33 = po.zernikes.zernike((3, 3), 'mn', r, theta)

    **References**

    [1] Vasudevan Lakshminarayanan and Andre Fleck (2011) -
    *Zernike polynomials: a guide*, Journal of Modern Optics
    
    [2] Robert J. Noll (1976) -
    "*Zernike polynomials and atmospheric turbulence*," J. Opt. Soc. Am. 66, 207-211

    """
    m, n = _to_mn(index, convention)
    return _zernike_mn(m, n, rho, theta, normalize, fill_value)


def _zernike_mn(m, n, rho, theta, normalize=True, fill_value=0):
    """Calculates zernike polynomial for given indices m, n for rho and theta.
    Implementation based on [1]. If normalize is `True` output is normalized
    following Noll's convention [2].

    :param m: zernike m-degree, angular frequency
    :type m: int
    :param n: zernike n-degree, radial order
    :type n: int
    :param rho: radial coordinates of the plane,
    :type rho: numpy.array
    :param theta: angle coordinate of the plane
    :type theta: numpy.array
    :param normalize: Should output be normalized following Noll's convention?
    :type normalize: bool
    :param fill_value: what value should be put for r > 1
    :type fill_value: numeric
    :return: resulting _zernike_mn distribution
    :rtype: numpy.array

    **References**

    [1] Vasudevan Lakshminarayanan and Andre Fleck (2011) -
    *Zernike polynomials: a guide*, Journal of Modern Optics

    [2] Robert J. Noll (1976) -
    "*Zernike polynomials and atmospheric turbulence*," J. Opt. Soc. Am. 66, 207-211

    """
    if n < 0:
        raise ValueError(f"Radial order n = {n} < 0! Must be n > 0.")
    if m > n:
        raise ValueError(f"Radial order n > m angular frequency: {n} > {m}.")
    if (n - m) % 2 != 0:
        raise ValueError(f"Radial order n - m angular frequency is not even: "
                         f"{n} - {m} = {n - m}")
    if isinstance(rho, u.Quantity):
        rho = rho.value
    if isinstance(theta, u.Quantity):
        theta = theta.value
    flat_disk = np.ones_like(rho)
    flat_disk[rho > 1] = fill_value
    radial_function = R(m, n, rho) * flat_disk
    if m >= 0:
        output = radial_function * np.cos(m * theta)
    else:
        output = radial_function * np.sin(-m * theta)

    if normalize:
        output *= norm_coefficient(m, n)

    return output


def R(m, n, rho):
    r"""The radial function described by:

    :math: `R_{n}^{m}(r)=\sum_{l=0}^{(n-m) / 2} \frac{(-1)^{l}(n-l) !}{l !\left[\frac{1}{2}(n+m)-l\right] !\left[\frac{1}{2}(n-m)-l\right] !} r^{n-2 l}`


    The definition was taken from [1], however, good description of zernikes
    can be found in almost all textbooks.

    :param m: m-degree, angular frequency
    :type m: int
    :param n: n-degree, radial order
    :type n: int
    :param rho: r coordinates of the plane,
    :type rho: numpy.array
    :return: calculated radial function
    :rtype: numpy.array

    **Example**

    >>> import astropy.units as u
    >>> import pyoptica as po
    >>>
    >>> npix = 101
    >>> pixel_scale = 0.5 * u.m
    >>> x, y = po.utils.mesh_grid(npix, pixel_scale)
    >>> r_max = .5 * npix * pixel_scale
    >>> r, _ = po.utils.cart2pol(x, y)
    >>> r = r / r_max
    >>> m, n =  1, 3
    >>> radial_poly = po.zernikes.R(m, n, r)

    **References**

    [1] Vasudevan Lakshminarayanan and Andre Fleck (2011) -
    *Zernike polynomials: a guide*, Journal of Modern Optics

    """

    m = abs(m)  # For Zm and Z-m, Rm is the same
    range_stop = (n - m) // 2 + 1
    output = np.zeros_like(rho)
    for l in range(range_stop):
        numerator = (-1.) ** l * factorial(n - l)
        denominator = factorial(l)
        denominator *= factorial((n + m) / 2. - l)
        denominator *= factorial((n - m) / 2. - l)
        output += numerator / denominator * rho ** (n - 2. * l)
    return output


def norm_coefficient(m, n):
    """Calculate normalizaiton coeficient of each zernikes; Rn+-m(1) = 1 for
    all n, m. THhe folowing euqtion is used (eq. 4 in [1]):

    :math:`N_{n}^{m}=\left(\frac{2(n+1)}{1+\delta_{m 0}}\right)^{1 / 2}`

    :param m: m-degree, angular frequency
    :type m: int
    :param n: n-degree, radial order
    :type n: int
    :return: calculate normalization coefficient
    :rtype: float

    **Example**

    >>> import pyoptica as po
    >>>
    >>> m, n = 1, 3
    >>> po.zernikes.norm_coefficient(m, n)
    2.8284271247461903

    **References**

    [1] Vasudevan Lakshminarayanan and Andre Fleck (2011) -
    *Zernike polynomials: a guide*, Journal of Modern Optics

    """
    numerator = 2 * (n + 1)
    denominator = 2 if m == 0 else 1
    norm = (numerator / denominator) ** 0.5
    return norm


def mn_to_osa(m, n):
    """Conversion of m, n indices to OSA j index [1]

    :param m: m-degree, angular frequency
    :type m: int
    :param n: n-degree, radial order
    :type n: int
    :return: j-index
    :rtype: int

    **Example**

    >>> import pyoptica as po
    >>>
    >>> m, n = 1, 3
    >>> po.zernikes.mn_to_osa(m, n)
    8

    **References**

    [1] Vasudevan Lakshminarayanan and Andre Fleck (2011) -
    *Zernike polynomials: a guide*, Journal of Modern Optics

    """
    j = (n * (n + 2) + m) // 2
    logger.debug(f"Converted (m={m}, n={n}) -> j={j}.")
    return j


def osa_to_mn(j):
    """Conversion of OSA j index to m, n [1]

    :param j: j index (OSA convention)
    :type j: int
    :return: m, n - angular frequency and radial order
    :rtype: Tuple(int, int)

    **Example**

    >>> import pyoptica as po
    >>>
    >>> j = 8
    >>> po.zernikes.osa_to_mn(j)
    (1, 3)

    **References**

    [1] Vasudevan Lakshminarayanan and Andre Fleck (2011) -
    *Zernike polynomials: a guide*, Journal of Modern Optics

    """
    n = ceil((-3 + sqrt((9 + 8 * j))) / 2)
    m = 2 * j - n * (n + 2)
    logger.debug(f"Converted j={j} -> (m={m}, n={n}).")
    return m, n


def mn_to_noll(m, n):
    """Conversion of m, n indices to Noll j index [1].

    :param m: m-degree, angular frequency
    :type m: int
    :param n: n-degree, radial order
    :type n: int
    :return: j-index
    :rtype: int

    **Example**

    >>> import pyoptica as po
    >>>
    >>> m, n = 1, 3
    >>> po.zernikes.mn_to_noll(m, n)
    8

    **References**

    [1] https://en.wikipedia.org/wiki/Zernike_polynomials

    """
    first = (n * (n + 1)) // 2
    second = abs(m)
    third = 0
    # I wanted to simplify the logic but it seems to work like a colander
    if m > 0 and n % 4 in [0, 1]:
        third = 0
    elif m < 0 and n % 4 in [2, 3]:
        third = 0
    elif m >= 0 and n % 4 in [2, 3]:
        third = 1
    elif m <= 0 and n % 4 in [0, 1]:
        third = 1
    j = first + second + third
    logger.debug(f"Converted (m={m}, n={n}) -> j={j}.")
    return j


def noll_to_mn(j):
    """Conversion of Noll j index to m, n [1]

    :param j: j index (Noll convention)
    :type j: int
    :return: m, n - angular frequency and radial order
    :rtype: Tuple(int, int)

    **Example**

    >>> import pyoptica as po
    >>>
    >>> j = 8
    >>> po.zernikes.noll_to_mn(j)
    (1, 3)

    **References**

    [1] Alfred K. K. Wong (2005) -
    *Optical Imaging in Projection Microlithography*, SPIE Press

    """
    n = ceil((-3 + sqrt((1 + 8 * j))) / 2)
    m = n - 2 * floor(((n + 2) * (n + 1) - 2 * j) / 4)
    if j & 1:  # is odd?
        m *= -1
    logger.debug(f"Converted j={j} -> (m={m}, n={n}).")
    return m, n


def fringe_to_mn(j):
    """Conversion of Fringe j index to m, n

    :param j: j index (Fringe convention)
    :type j: int
    :return: m, n - angular frequency and radial order
    :rtype: Tuple(int, int)

    **Example**

    >>> import pyoptica as po
    >>>
    >>> j = 8
    >>> po.zernikes.fringe_to_mn(j)
    (-1, 3)
    """
    # How did I get here? I couldn't get the equation anywhere on the Internet.
    # Let me explain you my train of thoughts. If you look closely at this:
    # https://www.spiedigitallibrary.org/ContentImages/Journals/JMMMGF/8/3/031404/WebImages/031404_1_2.jpg
    # You can clearly see that finding m (based on the row) is not a challenge
    # (just make sure that m is of correct sign). Once you know m just use
    # the equation from Wikipedia: all you have to do is solve a quadratic
    # equation given that n>=0.
    row = ceil(sqrt(j))
    m = (((row ** 2 - j) + 1) // 2)
    if (row ** 2 - j) % 2 != 0:
        m *= -1
    if m >= 0:
        n = int(2 * sqrt(2 * m + j) - m - 2)
    else:
        n = int(m - 2 + 2 * sqrt(j - 2 * m - 1))
    return m, n


def mn_to_fringe(m, n):
    """Conversion of Fringe j index to m, n [1]

    :param m: m-degree, angular frequency
    :type m: int
    :param n: n-degree, radial order
    :type n: int
    :return: j-index
    :rtype: int

    **Example**

    >>> import pyoptica as po
    >>>
    >>> m, n = -1, 3
    >>> po.zernikes.mn_to_fringe(m, n)
    8

    **References**

    [1] https://en.wikipedia.org/wiki/Zernike_polynomials
    """
    if m >= 0:
        j = (1 + (n + m) // 2) ** 2 - 2 * m
    else:
        j = (1 + (n - m) // 2) ** 2 + 2 * m + 1
    return j


def _to_mn(index, convention):
    """Helper function to converts given index (single) to mn indices (tuple
    of ints)

    :param index: index to be converted to m, n
    :type index: int or Tuple(int, int)
    :param convention: used indexing convention
    :type convention: str
    :return: m-angular frequency, n-radial order
    :rtype: Tuple(int, int)
    """
    _convention = convention.lower()
    if _convention == 'mn':
        m, n = index
    elif _convention == 'noll':
        m, n = noll_to_mn(index)
    elif _convention == 'osa':
        m, n = osa_to_mn(index)
    elif _convention == 'fringe':
        m, n = fringe_to_mn(index)
    else:
        raise ValueError(
            f'Zernike convention `{convention}` is not recognized! Only `mn`,'
            f' `osa`, `noll`, and `fringe` are available.'
        )
    return m, n


def construct_wavefront(
        index_coefs_dict, convention, rho, theta, normalize=True, fill_value=0
):
    """Sums up zernikes with given coefs to form the final wavefront:

    wf = sum(coef * zernike(index, convention, rho, theta) for index, coef in index_coefs_dict.items())

    :param index_coefs_dict: a dict with corresponding coefficients
    :type index_coefs_dict: dict
    :param convention: used indexing convention
    :type convention: str
    :param rho: r coordinates of the plane,
    :type rho: numpy.array
    :param theta: calculated radial function
    :type theta: numpy.array
    :param normalize: should wavefront be normalized?
    :type normalize: bool
    :param fill_value: what value should be put for r > 1
    :type fill_value: numeric
    :return: fitted wf (summed zernikes)
    :rtype: numpy.array

    **Example**

    >>> import astropy.units as u
    >>> import matplotlib.pyplot as plt
    >>> import numpy as np
    >>> import pyoptica as po
    >>> # Now we construct array which will be used for fitting
    >>> npix = 101
    >>> pixel_scale = 0.5 * u.m
    >>> x, y = po.utils.mesh_grid(npix, pixel_scale)
    >>> r_max = .5 * npix * pixel_scale
    >>> r, theta = po.utils.cart2pol(x, y)
    >>> r = r / r_max
    >>> r, theta = r.value, theta.value
    >>> # Now it is time to construct a wf (sum up zernikes):
    >>> coefs = [0.1] * 12
    >>> js = list(range(12))
    >>> index_coefs_dict = dict(zip(js, coefs))
    >>> convention = 'noll'
    >>> wf = po.zernikes.construct_wavefront(index_coefs_dict, convention, r, theta)

    """
    wf = np.zeros(rho.shape)
    for index, w in index_coefs_dict.items():
        wf += w * zernike(
            index, convention, rho, theta, normalize, fill_value=fill_value)
    return wf


def fit_zernikes(
        wf, indices, convention, rho, theta, normalize=True, cache=True,
        method='COBYLA', **kwargs):
    """Fits zernikes to the given wf -- for given `indices` finds corresponding
    coefficients `c_j` that when summed form fitted_wf:
        fitted_wf = sum((c_j * zernike_j(j, rho, theta, normalize))
    for which the difference (more specifically: l2-norm of the difference) is
    minimized.

    For the minimization routine `scipy.optimize.minimize` is used. For details
    please go to: https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html

    :param wf: a wavefront to be fitted
    :type wf: numpy.array
    :param indices: a list of zernikes to be fitted
    :type indices: iterable
    :param convention: used indexing convention
    :type convention: str
    :param rho: r coordinates of the plane,
    :type rho: numpy.array
    :param theta: calculated radial function
    :type theta: numpy.array
    :param normalize: should wavefront be normalized?
    :type normalize: bool
    :param cache: should zernikes be cached?
    :type cache: bool
    :param method: method of `scipy.optimize.minimize`
    :type method: str
    :param kwargs: remaining arguments for `scipy.optimize.minimize`
    :return: (found coefficients, minimization results)
    :rtype: Tuple(dict, OptimizeResult)

    **Example**

    >>> import astropy.units as u
    >>> import matplotlib.pyplot as plt
    >>> import numpy as np
    >>> # And from pyoptica
    >>> import pyoptica as po
    >>> # Now we construct array which will be used for fitting
    >>> npix = 101
    >>> pixel_scale = 0.5 * u.m
    >>> x, y = po.utils.mesh_grid(npix, pixel_scale)
    >>> r_max = .5 * npix * pixel_scale
    >>> r, theta = po.utils.cart2pol(x, y)
    >>> r = r / r_max
    >>> r, theta = r.value, theta.value
    >>> # Now it is time to construct a wf (sum up zernikes):
    >>> coefs = [0.1] * 12
    >>> indices = list(range(12))
    >>> js_coefs_dict = dict(zip(indices, coefs))
    >>> convention = 'noll'
    >>> wf = po.zernikes.construct_wavefront(js_coefs_dict, convention, r, theta)
    >>> fit, res = po.zernikes.fit_zernikes(wf, indices, convention, r, theta, cache=True)
    >>> fit
        {0: 0.09999999466400215,
         1: 0.09999998903721308,
         2: 0.09999999890069368,
         3: 0.09999999483945994,
         4: 0.09999999212142557,
         5: 0.0999999933147726,
         6: 0.0999999933135079,
         7: 0.09999999519388976,
         8: 0.09999999575002586,
         9: 0.0999999900420963,
         10: 0.09999999224797648,
         11: 0.0999999967091342}

    """
    if np.isnan(wf).any():
        raise ValueError(
            'Nan values are not accepted in the fitting routine. Please use'
            ' a different `fill_value` while constructing input wf!'
        )
    initial_coefs = np.zeros(len(indices))
    if cache:
        cached_zernikes = [
            zernike(i, convention, rho, theta, normalize) for i in indices]
        res = minimize(
            _l2_norm_of_aberration_wavefront_cache,
            x0=initial_coefs,
            args=(wf, cached_zernikes),
            method=method,
            **kwargs)
    else:
        res = minimize(
            _l2_norm_of_aberration_wavefront,
            x0=initial_coefs,
            args=(indices, convention, wf, rho, theta, normalize),
            method=method,
            **kwargs)
    return dict(zip(indices, res.x)), res


def fit_zernikes_lstsq(wf, indices, convention, rho, theta, normalize=True):
    """ Fits zernikes using the least squares method:

            Zi * ci = wf
    Where Zi and ci represent vectors with zernikes and coefficients
    respectively.

    This is the recommended method to fit a wavefront.

    :param wf: a wavefront to be fitted
    :type wf: numpy.array
    :param js: a list of zernikes to be fitted
    :type js: iterable
    :param convention: used indexing convention
    :type convention: str
    :param rho: r coordinates of the plane,
    :type rho: numpy.array
    :param theta: calculated radial function
    :type theta: numpy.array
    :param normalize: should wavefront be normalized?
    :type normalize: bool
    :return: fitted coeffs in a dict, (residuals, rank, singular_values)
    :rtype: Tuple(dict, tuple)

    **Example**

    >>> import astropy.units as u
    >>> import matplotlib.pyplot as plt
    >>> import numpy as np
    >>> # And from pyoptica
    >>> import pyoptica as po
    >>> # Now we construct array which will be used for fitting
    >>> npix = 101
    >>> pixel_scale = 0.5 * u.m
    >>> x, y = po.utils.mesh_grid(npix, pixel_scale)
    >>> r_max = .5 * npix * pixel_scale
    >>> r, theta = po.utils.cart2pol(x, y)
    >>> r = r / r_max
    >>> r, theta = r.value, theta.value
    >>> # Now it is time to construct a wf (sum up zernikes):
    >>> coefs = [0.1] * 12
    >>> indices = list(range(12))
    >>> js_coefs_dict = dict(zip(indices, coefs))
    >>> convention = 'noll'
    >>> wf = po.zernikes.construct_wavefront(js_coefs_dict, convention, r, theta)
    >>> fit, res = po.zernikes.fit_zernikes_lstsq(wf, indices, r, theta)
    >>> fit
        {0: 0.10000000000000002,
         1: 0.10000000000000016,
         2: 0.10000000000000019,
         3: 0.1,
         4: 0.09999999999999987,
         5: 0.10000000000000009,
         6: 0.09999999999999998,
         7: 0.10000000000000002,
         8: 0.10000000000000009,
         9: 0.10000000000000003,
         10: 0.1,
         11: 0.10000000000000003}

    """

    # It is not possible to work on 2D arrays... Therefore the need to be
    # flattened. That doesn't matter though -- it is just a way of data
    # representation.
    zernikes = np.array(
        [zernike(i, convention, rho, theta, normalize).flatten() for i in
         indices]
    ).T
    coefs, *res = np.linalg.lstsq(zernikes, wf.flatten(), rcond=None)
    return dict(zip(indices, coefs)), res


def _l2_norm_of_aberration_wavefront_cache(coefs, wf, cached_zernikes):
    """Calculates l2-norm between wavefront and a sum of zernikes in
    `cached_zernikes` and corresponding coefficients `coefs`.

    To be used in an optimization routine.

    :param coefs: a list of coefficients corresponding to zernikes
    :type coefs: iterable
    :param wf: a wavefront with which l2-norm is calculated
    :type wf: numpy.array
    :param cached_zernikes: a list of precalculated zernikes
    :type cached_zernikes: iterable
    :return: l2-norm of the difference between the two wavefronts
    :type: float

    """
    constructed_wf = np.zeros_like(wf)
    for z_dist, w in zip(cached_zernikes, coefs):
        constructed_wf += w * z_dist
    norm = np.linalg.norm(wf - constructed_wf)
    logger.debug(f"L2-norm = {norm}")
    return norm


def _l2_norm_of_aberration_wavefront(
        coefs, indices, convention, wavefront, rho, theta, normalize):
    """
    Calculates l2-norm between wavefront and a sum of zernikes of indices `js`
    and corresponding coefficients `coefs`.

    To be used in an optimization routine.

    :param coefs: a list of coefficients corresponding to zernikes
    :type coefs: iterable
    :param indices: a list of zernikes
    :type indices: iterable
    :param convention: used indexing convention
    :type convention: str
    :param wavefront: a wavefront with which l2-norm is calculated
    :type wavefront: numpy.array
    :param rho: r coordinates of the plane,
    :type rho: numpy.array
    :param theta: calculated radial function
    :type theta: numpy.array
    :param normalize: should wavefront be normalized?
    :type normalize: bool
    :return: l2-norm of the difference between the two wavefronts
    :type: float

    """
    index_coef = dict(zip(indices, coefs))
    sum_wf = construct_wavefront(index_coef, convention, rho, theta, normalize)
    norm = np.linalg.norm(wavefront - sum_wf)
    logger.debug(f"L2-norm = {norm}")
    return norm


def get_zernike_name(index, convention, latex=True):
    """Gets name of the zernike for given convention for index < 17 (Fringe)
    returns also the name.

    :param index: index of zernike in given convention
    :type index: int or Tuple(int, int)
    :param convention: used indexing convention
    :type convention: str
    :param latex: add latex equation formatting?
    :type latex: bool
    :return: name of zernike
    :rtype: str

    **Example**

    >>> import pyoptica as po
    >>> po.zernikes.get_zernike_name(7, 'fringe')
    '$Z_{7}$ x-Coma'
    >>> po.zernikes.get_zernike_name(7, 'fringe', latex=False)
    'Z_7 x-Coma'
    """
    _convention = convention.lower()
    m, n = _to_mn(index, convention)
    if _convention in ['noll', 'osa', 'fringe']:
        if latex:
            name = f'$Z_{{{index}}}$'
        else:
            name = f'Z_{index}'
    else:
        if latex:
            name = f'$Z_{{{n}}}^{{{m}}}$'
        else:
            name = f'Z_{n}^{m}'

    if (m, n) in ZERNIKE_NAMES:
        name += ' ' + ZERNIKE_NAMES[(m, n)]
    return name



from . import base_optical_element
from . import binary_grating
from . import circular_reticle
from . import diffuser
from . import free_space
from . import rectangular_reticle
from . import siemens_star
from . import thin_lens
from .base_optical_element import BaseOpticalElement
from .binary_grating import BinaryGrating
from .circular_reticle import CircularAperture, CircularObscuration
from .diffuser import Diffuser
from .free_space import FreeSpace
from .rectangular_reticle import RectangularAperture, RectangularObscuration
from .siemens_star import SiemensStar
from .thin_lens import ThinLens

__all__ = ['base_optical_element', 'BaseOpticalElement',
           'circular_reticle', 'CircularAperture', 'CircularObscuration',
           'diffuser', 'Diffuser',
           'rectangular_reticle', 'RectangularAperture', 'RectangularObscuration',
           'thin_lens', 'ThinLens',
           'free_space', 'FreeSpace',
           'binary_grating', 'BinaryGrating',
           'siemens_star', 'SiemensStar'
           ]



import abc
import copy
import pickle

import astropy.units as u
import numpy as np

from . import logging
from . import plotting
from . import utils


class Distribution(logging.LoggerMixin, plotting.Plottable):
    """Abstract class representing a 2D distribution. """

    @u.quantity_input(wavelength=u.m, pixel_scale=u.m)
    def __init__(self, wavelength, pixel_scale, npix):
        self.wavelength = wavelength
        self.pixel_scale = pixel_scale
        self.npix = npix
        self.size = npix * pixel_scale
        self.k = np.pi * 2.0 / self.wavelength
        self.x, self.y = utils.mesh_grid(npix, pixel_scale)

    def __repr__(self):
        return f"{self.__class__.__name__} " \
            f"of wavelength = {self.wavelength:.2f}, " \
            f"pixel scale = {self.pixel_scale:.2f}, " \
            f"size {self.npix} x {self.npix}."

    def copy(self):
        """A method to *deep* copy the distribution.

        :return: a copy of the wavefront
        :rtype: pyoptica.Wavefront

        **Example**

        >>> import astropy.units as u
        >>> import pyoptica as po
        >>>
        >>> wavelength = 500 * u.nm
        >>> pixel_scale = 22 * u.um
        >>> npix = 1024
        # As an example we use a Wavefront. Distribution is an abstractclass!
        >>> wf = po.Wavefront(wavelength, pixel_scale, npix)
        >>> new_wf = wf.copy()

        """
        return copy.deepcopy(self)

    def to_pickle(self, path):
        """A method to serialize distribution to pickle.

        :param path: path to the file to which distribution is dumped
        :type path: str

        **Example**

        >>> import astropy.units as u
        >>> import pyoptica as po
        >>>
        >>> wavelength = 500 * u.nm
        >>> pixel_scale = 22 * u.um
        >>> npix = 1024
        >>> # We are going to work on `Image` as `Distribution` is abstract
        >>> image = Image(wavelength, pixel_scale, npix)
        >>> file_path = 'path/to/your/file.pickle'
        >>> image.to_pickle(file_path)

        """
        with open(path, 'wb') as file:
            pickle.dump(self, file, pickle.HIGHEST_PROTOCOL)
        self.logger.debug(f"Saved {self.__class__.__name__} to {path}")

    @classmethod
    def from_pickle(cls, path):
        """A method to load distribution serialized to pickle.

        :param path: path to the file to be loaded.
        :type path: str
        :returns: loaded distribution
        :rtype: pyoptica.Distribution (or any of the subclasses)

        **Example**

        >>> import pyoptica as po
        >>>
        >>> file_path = 'path/to/your/file.pickle'
        >>> wf = po.Wavefront.from_pickle(file_path)

        """
        with open(path, 'rb') as file:
            wf = pickle.load(file)
        cls.logger.debug(f"Loaded {cls.__class__.__name__} from {path}")
        return wf

    @abc.abstractmethod
    def to_fits(self, path):
        pass

    @classmethod
    @abc.abstractmethod
    def from_fits(cls, path):
        pass



import logging
import warnings

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
import pkg_resources

_log = logging.getLogger(__name__)
_log.setLevel(logging.DEBUG)
_ch = logging.StreamHandler()
_ch.setLevel(logging.DEBUG)
_formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
_ch.setFormatter(_formatter)
_log.addHandler(_ch)

LOGOS = dict(
    round=pkg_resources.resource_filename(
        __name__, 'data/logos/pyoptica_round.png'),
    rect=pkg_resources.resource_filename(
        __name__, 'data/logos/pyoptica_rect.png'),
    text=pkg_resources.resource_filename(
        __name__, 'data/logos/pyoptica_text_only.png'),
)


def cart2pol(x, y):
    """Calculates polar coordinates from given cartesian coordinates.

    :param x: X coordinate
    :type x: float
    :param y: Y coordinate
    :type y: float
    :return: Polar coordinates of given point
    :rtype: tuple[float, float]

    **Example**

    >>> import pyoptica as po
    >>>
    >>> x, y = 1 * u.m, 1 * u.m
    >>> po.utils.cart2pol(x, y)
    (<Quantity 1.41421356 m>, <Quantity 0.78539816 rad>)

    """

    rho = np.hypot(x, y)
    phi = np.arctan2(y, x)
    return rho, phi


def pol2cart(rho, phi):
    """Calculates cartesian coordinates from given polar coordinates.

    :param rho: Radial coordinate
    :type rho: float
    :param phi: Angular coordinate
    :type phi: float
    :return: Cartesian coordinates of given point
    :rtype: tuple[float, float]

    **Example**

    >>> import pyoptica as po
    >>>
    >>> r, p = 1 * u.m, 1 * u.rad
    >>> po.utils.pol2cart(r, p)
    (<Quantity 0.54030231 m>, <Quantity 0.84147098 m>)
    """

    x = rho * np.cos(phi)
    y = rho * np.sin(phi)
    return x, y


def fft(array, method='numpy'):
    """Performs 2D FFT for array with given method

    :param array: array to be Fourier transformed
    :type array: np.array
    :param method: description of algorithm (package)
    :type method: str
    :return: Fourier transformed array
    :rtype: np.array

    **Example**

    >>> import pyoptica as po
    >>>
    >>> arr = np.ones((1024, 1024))
    >>> po.utils.fft(arr)

    """
    if method == 'numpy':
        ishifted_array = np.fft.ifftshift(array)
        fft_array = np.fft.fft2(ishifted_array)
        shifted_fft_array = np.fft.fftshift(fft_array)
        return shifted_fft_array
    else:
        raise ValueError(f"")


def ifft(array: np.array, method='numpy') -> np.array:
    """Performs 2D IFFT for array with given method

    :param array: array to be inversely Fourier transformed
    :type array: np.array
    :param method: description of algorithm (package)
    :type method: str
    :return: Fourier transformed array
    :rtype: np.array

    **Example**

    >>> import pyoptica as po
    >>>
    >>> arr = np.ones((1024, 1024))
    >>> po.utils.ifft(arr)

    """
    if method == 'numpy':
        ishifted_array = np.fft.ifftshift(array)
        ifft_array = np.fft.ifft2(ishifted_array)
        shifted_ifft_array = np.fft.fftshift(ifft_array)
        return shifted_ifft_array
    else:
        raise ValueError(f"")


def mesh_grid(npix, pixel_scale):
    """Create a grid of size `npix` with sampling defined by `pixel_scale`
    centered at 0,0.

    :param npix: Size of the grid (in each direction).
    :type npix: int
    :param pixel_scale: Sampling of the grid
    :type pixel_scale: astropy.Quantity (type is not forced!)
    :return: two dimensional mesh grid (x, y)
    :rtype: (np.array, np.array)

    **Example**

    >>> import astropy.units as u
    >>> import pyoptica as po
    >>>
    >>> pixel_scale = 1 * u.um
    >>> npix = 512
    >>> x, y = po.utils.mesh_grid(npix, pixel_scale)

    """
    y, x = (np.indices((npix, npix)) - npix / 2) * pixel_scale
    return x, y


def space2freq(x):
    """Converts space coordinates into frequency, x->f, y->g
    :param x: X grid
    :type x: astropy.Quantity array of type length
    :param y: Y grid
    :type y: astropy.Quantity array of type length
    :return: (array, array)

    """
    pixel_scale = x[0, 0] - x[1, 1]
    npix = x.shape[0]
    denominator = pixel_scale ** 2 * npix
    f = x / denominator
    return f


def range_validator(value, range):
    """Checks if value is within specified range and clamps it according to
    given boundaries

    :param value: Numeric value to check if it's range is valid.
    :type value: float
    :param range: Tuple with lower and upper limit for checked value.
    :type range: (float, float)
    :return: Given value or clamped to the given range.
    :rtype: float

    """
    min_val, max_val = range
    if value < min_val:
        warning = f"Value out of valid range: {range}, return value set to: {min_val}"
        warnings.warn(warning)
        _log.warning(warning)
        return min_val
    elif value > max_val:
        warning = f"Value out of valid range: {range}, return value set to: {max_val}"
        warnings.warn(warning)
        _log.warning(warning)
        return max_val
    else:
        return value


def param_validator(param, params):
    """Checks if parameter is listed in accepted parameters list, if not returns
    first listed valid parameter.

    :param param: Parameter to be check if is listed in valid parameters list.
    :type param: object
    :param params: List of valid parameters
    :type params: list of objects
    :return: Given parameter or first listed as valid.
    :rtype: float

    """
    if param not in params:
        warning = f"Parameter: {param} is not listed as valid: {params}, param set to default: {params[0]}"
        warnings.warn(warning)
        _log.warning(warning)
        return params[0]
    else:
        return param


def rgb2gray(rgb):
    """Converts rgb(a) image to grayscale. gray = 0.299r + 0.587g + 0.114b
    Based on: https://en.wikipedia.org/wiki/Grayscale

    :param rgb: array representing an image rgb(a) of size (n, m, 3/(4))
    :type: np.array
    :return: grayscale image of size (n, m)
    :rtype: np.array

    **Example**

    >>> import pyoptica as po
    >>> rgb = np.ones((16, 16, 3))
    >>> po.utils.rgb2gray(rgb)
    """
    return np.dot(rgb[..., :3], [0.299, 0.587, 0.114])


def get_logo(kind='round'):
    """Loads PyOptica logo as a numpy array. `rect`, `round`, and `text` kinds
    are available.

    :param kind: logo kind (`rect`, `round`, `text`)
    :type kind: str
    :return: loaded logo
    :rtype: np.array
    :raises: KeyError

    **Example**
    >>> import pyoptica as po
    >>> po.utils.get_logo()
    """
    try:
        path = LOGOS[kind]
    except KeyError as e:
        raise KeyError(f'{kind} is not supported.'
                       f' Only {list(LOGOS.keys())} are supported') from e
    img = plt.imread(path)
    img_gray = rgb2gray(img)
    return img_gray



import astropy.units as u
import numpy as np
from scipy.ndimage import rotate

from .base_optical_element import BaseOpticalElement
from .. import utils


class BinaryGrating(BaseOpticalElement):
    """ A class representing a binary amplitude/phase grating.

    :param period: Period of the grating.
    :type period: astropy.Quantity of type length
    :param duty: Duty cycle of the grating [0-1].
    :type duty: float
    :param angle: Angle of the grating in degrees.
    :type angle: astropy.Quantity of type angle
    :param kind: Type of grating ["amplitude", "phase"].
    :type kind: str
    :param a_trans: Transmission of the grating [0-1].
    :type a_trans: float
    :param o_thick: Optical thickness of the grating.
    :type o_thick: float


    **Example**

    >>> import astropy.units as u
    >>> import pyoptica as po
    >>>
    >>> period = 1 * u.mm
    >>> duty = 0.5
    >>> angle = 45 * u.deg
    >>> grating = po.BinaryGrating(period, duty, angle)

    """

    VALID_RANGE_DUTY = (0, 1)
    VALID_RANGE_ANGLE = (0 * u.deg, 360 * u.deg)
    VALID_RANGE_A_TRANS = (0, 1)
    VALID_RANGE_O_THICK = (0, 2 * np.pi)
    VALID_KIND = ["amplitude", "phase"]

    def __init__(self, period, duty=.5, angle=0 * u.deg, kind='amplitude', a_trans=1, o_thick=np.pi):
        self.period = period
        self.duty = utils.range_validator(duty, self.VALID_RANGE_DUTY)
        self.angle = utils.range_validator(angle, self.VALID_RANGE_ANGLE)
        self.kind = utils.param_validator(kind, self.VALID_KIND)
        self.a_trans = utils.range_validator(a_trans, self.VALID_RANGE_A_TRANS)
        self.o_thick = utils.range_validator(o_thick, self.VALID_RANGE_O_THICK)

    @classmethod
    def etch_grating(cls, npix, period, duty, angle=0):

        def _crop_center(img, cropx, cropy):
            y, x = img.shape
            startx = x // 2 - (cropx // 2)
            starty = y // 2 - (cropy // 2)
            return img[starty:starty + cropy, startx:startx + cropx]

        cell = [1] * int(period * duty) + [0] * int(period * (1 - duty))
        if angle:
            pad = np.sqrt(2)
        else:
            pad = 1
        profile_r = np.hstack([cell] * (int(pad * npix / period * .5 + period)))[:int(npix * pad / 2)]
        profile_l = np.hstack([cell[::-1]] * (int(pad * npix / period * .5 + period)))[:int(npix * pad / 2)][::-1]
        profile = np.hstack([profile_l, profile_r])
        grating = np.vstack([profile] * int(npix * pad))
        if angle:
            return _crop_center(rotate(grating, angle=angle, reshape=True), npix, npix)
        else:
            return grating

    def amplitude_transmittance(self, wavefront):
        if "amplitude" in self.kind:
            npix = wavefront.npix
            pix_period = (self.period.to(u.um).value // wavefront.pixel_scale.to(u.um).value)
            self.logger.info(f"Resulting period due to canvas granularity: "
                             f"{(pix_period * wavefront.pixel_scale).to(u.um)}"
                             )
            gr = self.etch_grating(npix, pix_period, self.duty, self.angle.to(u.deg).value) * self.a_trans
            return gr
        else:
            return np.ones_like(wavefront.amplitude)

    def phase_transmittance(self, wavefront):
        if "phase" in self.kind:
            npix = wavefront.npix
            pix_period = (self.period.to(u.um).value // wavefront.pixel_scale.to(u.um).value)
            self.logger.info(f"Resulting period due to canvas granularity: "
                             f"{(pix_period * wavefront.pixel_scale).to(u.um)}"
                             )
            gr = self.etch_grating(npix, pix_period, self.duty, self.angle.to(u.deg).value) * self.a_trans
            return np.exp(1.j * self.o_thick * gr)
        else:
            return np.exp(1.j * np.zeros_like(wavefront.phase))



import abc

from .. import logging
from ..wavefront import Wavefront


class BaseOpticalElement(logging.LoggerMixin, abc.ABC):
    """ An abstract class for optical elements. It is the base class for all
    future optical elements (e.g. a lens). All classes derived from this base
    one should implement `amplitude_transmittance` and `phase_transmittance`.
    They describe respectively how the element modulates amplitude and how it
    shifts phase of the wavefront the element interacts with.

    """

    @abc.abstractmethod
    def amplitude_transmittance(self, wavefront):
        pass

    @abc.abstractmethod
    def phase_transmittance(self, wavefront):
        pass

    def transmittance(self, wavefront):
        """Produces transmittance of the optical element"""
        t_amplitude = self.amplitude_transmittance(wavefront)
        t_phase = self.phase_transmittance(wavefront)
        t = t_amplitude * t_phase
        return t

    def __mul__(self, wavefront):
        """Multiplication with a wavefront.

        :param wavefront: Main pyoptica `Wavefront` class.
        :type wavefront: pyoptica.Wavefront
        :return: multiplied wavefront.
        :rtype: pyoptica.Wavefront
        """

        multiplied_wf = Wavefront(wavefront.wavelength, wavefront.pixel_scale, wavefront.npix)
        multiplied_wf.wavefront = wavefront.wavefront * self.transmittance(wavefront)
        return multiplied_wf

    __rmul__ = __mul__



import numpy as np

from .base_optical_element import BaseOpticalElement


class Diffuser(BaseOpticalElement):
    """ A class representing an optical diffuser.

    :param distribution: Distribution of phase shift ['uniform', 'normal'].
    :type distribution: str
    :param rotating: If *True* phase changes from call to call.
    :type rotating: bool

    **Example**

    >>> import astropy.units as u
    >>> import pyoptica as po
    >>>
    >>> distribution = 'normal'
    >>> rotating = True
    >>> diff = po.Diffuser(distribution)

    **References**

    [1] Joseph W. Goodman (2004) -
    *Introduction to Fourier Optics*, W. H. Freeman

    """
    DISTRIBUTIONS = ['uniform', 'normal']

    def __init__(self, distribution='uniform', rotating=False):

        if distribution not in self.DISTRIBUTIONS:
            raise ValueError(f"Valid distributions are: {self.DISTRIBUTIONS}")

        self.distribution = distribution
        self.rotating = rotating
        self._rnd_state = None
        if distribution == 'uniform':
            self._distribution_generator = np.random.rand
        elif distribution == 'normal':
            self._distribution_generator = np.random.randn
        self._seed = np.random.randint(0, 2**16-1)

    def _reset_seed(self):
        if not self.rotating:
            self._rnd_state = np.random.RandomState(self._seed)
            if self.distribution == 'uniform':
                self._distribution_generator = self._rnd_state.rand
            if self.distribution == 'normal':
                self._distribution_generator = self._rnd_state.randn

    def amplitude_transmittance(self, wavefront):
        """
        We assumed that the diffuser's amplitude transmittance is one .

        :param wavefront: The wavefront which interacts with the diffuser
        :type wavefront: pyoptica.Wavefront
        :return: distribution of amplitude transmittance resulting from the diffuser
        :rtype: numpy.array
        """

        return np.ones_like(wavefront.amplitude)

    def phase_transmittance(self, wavefront):
        """
        We assumed that the diffuser introduces completely random phase.

        :param wavefront: The wavefront which interacts with the diffuser
        :type wavefront: pyoptica.Wavefront
        :return: distribution of amplitude transmittance resulting from the diffuser
        :rtype: numpy.array
        """
        if not self.rotating:
            self._reset_seed()
        return np.exp(1.j * 2 * np.pi * (self._distribution_generator(*wavefront.phase.shape) - .5))



import astropy.units as u
import numpy as np

from . import logging
from . import plotting
from . import utils
from . import zernikes
from .image import Image
from .wavefront import Wavefront


class ImagingSystem(logging.LoggerMixin, plotting.Plottable):
    """ A class representing an imaging systems. It is based on reasoning
    presented in [1] and [2] (numerical implementation suggested in
    Chapter 7 [3]). An optical system composed of many optical elements can be
    described by one function that _aggregates_ all properties of them all.
    In case of coherent imaging the, the image is found based on
    (eqs. 5.37-5.40 in [1], the naming convention follows that given in
    [1] *Introduction to Fourier Optics* by J.W. Goodman)

    It is assumed that the system is isoplanatic (the aberrations are constant
    through out the field)

    *THE IMAGING SYSTEM HAS A CIRCULAR APERTURE!*

    :param wavelength: Wavelength used in simulations
    :type wavelength: astropy.Quantity of type length
    :param pixel_scale: Sampling of the wavefront
    :type pixel_scale: astropy.Quantity of type 1/length
    :param npix: Number of pixels in the wavefront array
    :type npix: int
    :param coherence_factor: complex coherence factor
    :type coherence_factor: float
    :param m: transverse magnification of the optical system
    :type m: float
    :param na: numerical aperture defined as :math:`NA = sin\theta_{max}`
    :type na: float
    :param n_o: refractive index on the object side
    :type n_o: float
    :param n_i: refractive index on the image side
    :type n_i: float

    **Example**

    >>> import astropy.units as u
    >>> import pyoptica as po
    >>>
    >>> wavelength = 1 * u.um
    >>> pixel_scale = 10 * u.mm
    >>> npix = 1024
    >>> coh_factor = 1  # Fully coherent
    >>> img_sys = po.ImagingSystem(wavelength, pixel_scale, npix, coh_factor)

    **References**

    [1] Joseph W. Goodman (2004) -
    *Introduction to Fourier Optics*, W. H. Freeman

    [2] Kedar Khare (2016) -
    *Fourier Optics and Computational Imaging*, Wiley&Sons Ltd.

    [3] David Voelz (2011) -
    *Computational Fourier Optics Matlab Tutorial*, Spie Press

    [4] M. Born, E. Wolf (1980) -
    *Principles of Optics*, Pergamon Press (Oxford UK)

    """
    PLOTTING_OPTIONS = {
        'psf': dict(
            x='x', y='y', title='|PSF|', z_function=lambda x: np.abs(x)
        ),
        'atf': dict(
            x='f', y='g', title='|ATF|', z_function=lambda x: np.abs(x)
        ),
        'ptf': dict(x='f', y='g', title='PTF'),
        'mtf': dict(x='f', y='g', title='MTF'),
        'otf': dict(
            x='f', y='g', title='|OTF|', z_function=lambda x: np.abs(x)
        ),
        'wavefront': dict(
            colorbar_title='[rad]', vmin=-np.pi, vmax=np.pi,
            bar_ticks=[-np.pi, -np.pi / 2, 0, np.pi / 2, np.pi],
            bar_ticks_labels=[
                r'$-\pi$',
                r'$-\frac{\pi}{2}$',
                r'$0$',
                r'$\frac{\pi}{2}$',
                r'$\pi$'
            ],
            x='f', y='g', cmap='hsv', title='Wavefront (aberrations)',
            z_function='_aberrations_plot_special'
        )
    }

    @u.quantity_input(wavelength=u.m, pixel_scale=u.m)
    def __init__(self, wavelength, pixel_scale, npix, coherence_factor, m=1,
                 na=1, n_o=1, n_i=1):
        self.wavelength = wavelength
        self.pixel_scale = pixel_scale
        self.npix = npix
        self.coherence_factor = coherence_factor
        self.m = m
        self.na = na
        self.n_o = n_o
        self.n_i = n_i

        self.x, self.y = utils.mesh_grid(npix, pixel_scale)

        self._atf = np.ones((npix, npix))
        self._otf = np.ones((npix, npix))

        self._zernike_coefs = dict()
        self._zernike_convention = None
        self._apodization = None
        self._system_modified = True

    def __setattr__(self, key, value):
        # I really like this!
        if key != '_system_modified':
            self._system_modified = True
        super().__setattr__(key, value)

    @property
    def f(self):
        f = utils.space2freq(self.x)
        return f

    @property
    def g(self):
        g = utils.space2freq(self.y)
        return g

    def load_zernikes(self, zernike_coefs, convention):
        """
        Loads zernike coefficients represented as a dict: `{index: val}`.
        The coefficients are expected to be in RADIANS!

        :param zernike_coefs: zernikes to be loaded
        :type zernike_coefs: dict
        :param convention: the way zernikes are represented (`mn`, `osa`, or `noll`)
        :type convention: str
        :raises: ValueError

        **Example**

        >>> import astropy.units as u
        >>> import pyoptica as po
        >>>
        >>> wl = 1 * u.um
        >>> pixel_scale = 10 * u.mm
        >>> npix = 1024
        >>> coh_factor = 1  # Fully coherent
        >>> img_sys = po.ImagingSystem(wl, pixel_scale, npix, coh_factor)
        >>>
        >>> zernike_coefs = {1: 0.2, 2: 0.3, 7: 0.1, 10: 0.01}  # OSA
        >>> img_sys.load_zernikes(zernike_coefs)
        """
        self._zernike_coefs = zernike_coefs
        self._zernike_convention = convention


    def _calc_atf(self):
        """Calculates Amplitude Transfer Function H based on the current state
        of the object (numerical aperture, magnification, aberrations, and
        apodization).

        """
        H_ideal = self._calc_diffraction_limited_atf()
        H = self._apply_radiometric_correction(H_ideal)
        H_with_aberrations = self._apply_aberrations(H)
        self._atf = H_with_aberrations

    def _calc_otf(self):
        """ Calculates Optical Transfer Function based on the current state
        of the object. It is calculated as a normalized autocorrelation of the
        Amplitude Transfer Function.
        """
        H_fft = utils.fft(self._atf)
        H_fft_abs_squared = np.abs(H_fft) ** 2
        otf = utils.ifft(H_fft_abs_squared)
        i, j = otf.shape
        otf /= otf[i // 2, j // 2]
        rho = np.hypot(self.f, self.g)
        otf[rho > 2 * self.na / self.wavelength] = 0
        self._otf = otf

    def calculate(self):
        """Precalculates both Amplitude and Optical Transfer Functions
        (including aberrations, magnification, apodization).

        This method must be called always after manipulating the imaging
        system otherwise the results will be inaccurate.

        **Example**

        >>> import astropy.units as u
        >>> import pyoptica as po
        >>>
        >>> wavelength = 1 * u.um
        >>> pixel_scale = 10 * u.mm
        >>> npix = 1024
        >>> mu = 1
        >>> img_sys = po.ImagingSystem(wavelength, pixel_scale, npix, mu)
        >>> img_sys.calculate()
        """
        self._calc_atf()
        self._calc_otf()
        self._system_modified = False

    @property
    def atf(self):
        """Amplitude Transfer Function (Goodman naming convention).

        **Example**

        >>> import astropy.units as u
        >>> import pyoptica as po
        >>>
        >>> wl = 1 * u.um
        >>> pixel_scale = 10 * u.mm
        >>> npix = 1024
        >>> coh_factor = 1
        >>> img_sys = po.ImagingSystem(wl, pixel_scale, npix, coh_factor)
        >>> img_sys.calculate()  # System must be precalculated first!
        >>> img_sys.atf

        """
        self._warn_if_modified()
        return self._atf

    @property
    def wavefront(self):
        """Wavefront (aberrations) of the imaging system.

        **Example**

        >>> import astropy.units as u
        >>> import pyoptica as po
        >>>
        >>> wl = 1 * u.um
        >>> pixel_scale = 10 * u.mm
        >>> npix = 1024
        >>> coh_factor = 1
        >>> img_sys = po.ImagingSystem(wl, pixel_scale, npix, coh_factor)
        >>> img_sys.calculate()  # System must be precalculated first!
        >>> img_sys.wavefront

        """
        self._warn_if_modified()
        return np.angle(self._atf)

    @property
    def otf(self):
        """Optical Transfer Function (Goodman naming convention).

        **Example**

        >>> import astropy.units as u
        >>> import pyoptica as po
        >>>
        >>> wl = 1 * u.um
        >>> pixel_scale = 10 * u.mm
        >>> npix = 1024
        >>> coh_factor = 1
        >>> img_sys = po.ImagingSystem(wl, pixel_scale, npix, coh_factor)
        >>> img_sys.calculate()  # System must be precalculated first!
        >>> img_sys.otf

        """
        self._warn_if_modified()
        return self._otf

    @property
    def psf(self):
        """Amplitude Point-Spread Function (Goodman naming
        convention).

        **Example**

        >>> import astropy.units as u
        >>> import pyoptica as po
        >>>
        >>> wl = 1 * u.um
        >>> pixel_scale = 10 * u.mm
        >>> npix = 1024
        >>> coh_factor = 1
        >>> img_sys = po.ImagingSystem(wl, pixel_scale, npix, coh_factor)
        >>> img_sys.calculate()  # System must be precalculated first!
        >>> img_sys.psf

        """
        return utils.ifft(self.atf)

    @property
    def mtf(self):
        """Modulation Transfer Function (MTF = |OTF|)
         (Goodman naming convention).

        **Example**

        >>> import astropy.units as u
        >>> import pyoptica as po
        >>>
        >>> wl = 1 * u.um
        >>> pixel_scale = 10 * u.mm
        >>> npix = 1024
        >>> coh_factor = 1
        >>> img_sys = po.ImagingSystem(wl, pixel_scale, npix, coh_factor)
        >>> img_sys.calculate()  # System must be precalculated first!
        >>> img_sys.mtf
        """
        return np.abs(self.otf)

    @property
    def ptf(self):
        """Phase Transfer Function (PTF = arg(OTF))
         (Goodman naming convention).

        **Example**

        >>> import astropy.units as u
        >>> import pyoptica as po
        >>>
        >>> wl = 1 * u.um
        >>> pixel_scale = 10 * u.mm
        >>> npix = 1024
        >>> coh_factor = 1
        >>> img_sys = po.ImagingSystem(wl, pixel_scale, npix, coh_factor)
        >>> img_sys.calculate()  # System must be precalculated first!
        >>> img_sys.ptf
        """
        return np.angle(self.otf)

    def _warn_if_modified(self):
        if self._system_modified:
            self.logger.warning(
                'The system has been modified, however, it has not been '
                'recalculated. The results may be inaccurate. '
                'Please rerun `calculate` to obtain results that represent '
                'the current state of the system.')

    def _calc_diffraction_limited_atf(self):
        """
        The diffraction limited Amplitude Transfer Function is calculated:

        :math:`\left\{\begin{matrix} 1, & \text{if } f^2 +g^2 < f_0^2 \\ 0, & \text{otherwise} \end{matrix}\right.`

        :return: Diffraction limited Amplitude Transfer Function
        :rtype: np.array

        """
        rho_squared = self.f ** 2 + self.g ** 2
        H = np.where(rho_squared < (self.na / self.wavelength) ** 2, 1, 0)
        return H

    def _apply_radiometric_correction(self, atf_ideal):
        """ If there is magnification (or reduction) in the system, the
        distribution of light entering the system will be different than the
        distribution leaving it. In order to include that in the Amplitude
        Transfer Function H, a radiometric correction has to be applied
        (following M. Born and E. Wolf [4]):

        :math:`\left ( \frac{1 - \lambda^2(f^2 +g^2)M^2}{1 - \lambda^2(f^2 +g^2)}\right )^{0.25}`

        :param atf_ideal: diffraction limited Amplitude Transfer Function
        :type atf_ideal: np.array
        :return: Amplitude Transfer Function with radiometric correction
        :rtype: np.array

        """
        rho_squared = self.f ** 2 + self.g ** 2
        numerator = 1 - (
                self.wavelength ** 2 * rho_squared) * self.m ** 2 / self.n_o ** 2
        denominator = 1 - (self.wavelength ** 2 * rho_squared) / self.n_i ** 2
        r = numerator / denominator
        H = np.power(atf_ideal * r, 0.25)
        H = H.to(u.dimensionless_unscaled).value
        #  Maybe it would make more sense to raise an exception if NA > n
        H[~np.isfinite(H)] = 0
        return H

    def _apply_aberrations(self, H):
        """Apply aberrations from `self._zernike_coefs` to the amplitude transfer
        function (H):

        :math:`H=H_d \cdot \exp(2 \pi i W)`

        Where `H_d` is the diffraction limited transfer function, and `W` the
        calculated wavefront.

        :param H: The diffraction limited H
        :type H: np.array
        :return: amplitude transfer function with aberrations
        :rtype: np.array

        """
        rho, theta = utils.cart2pol(self.f, self.g)
        rho = rho.to(1 / u.m) / (self.na / self.wavelength).to(1 / u.m)
        wf = zernikes.construct_wavefront(
            self._zernike_coefs, self._zernike_convention, rho, theta
        )
        return H * np.exp(2.j * np.pi * wf)

    def image_wavefront(self, wf):
        """ A method to image given wavefront. There are three distinct cases:

        +--------------------+------------------------------+---------------------------------+
        | Coherence          | Fundamental Quantity         | Transfer Function               |
        +====================+==============================+=================================+
        | Fully coherent     | Input wavefront field        | Amplitude Transfer Function     |
        +--------------------+------------------------------+---------------------------------+
        | Fully incoherent   | Intensity of input wavefront | Optical Transfer Function       |
        +--------------------+------------------------------+---------------------------------+
        | Partially coherent | Mutual intensity             | Transmission Cross-Coefficients |
        +--------------------+------------------------------+---------------------------------+

        :param wf: Input wavefront
        :type wf: pyoptica.Wavefront
        :return: image in the image plane
        :rtype: pyoptica.Image

        **Example**

        >>> import astropy.units as u
        >>> import pyoptica as po
        >>>
        >>> wl = 1 * u.um
        >>> pixel_scale = 10 * u.mm
        >>> npix = 1024
        >>> coh_factor = 1  # Fully coherent
        >>> img_sys = ImagingSystem(wl, pixel_scale, npix, coh_factor)
        >>> img_sys.calculate()  # Remember to call `calculate`!
        >>> wf = Wavefront(wl, 1/pixel_scale/npix, npix)
        >>> imaged_wf = img_sys.image_wavefront(wf)
        >>> imaged_wf

        """
        self._check_wavefront_compatibility(wf)

        if self.coherence_factor == 1:
            return self._image_wavefront_fully_coherent(wf)
        elif self.coherence_factor == 0:
            return self._image_wavefront_fully_incoherent(wf)
        elif 0 < self.coherence_factor < 1:
            return self._image_wavefront_partially_coherent(wf)
        else:
            raise ValueError('Complex coherence factor is limited to [0, 1]!')

    def _image_wavefront_fully_coherent(self, wf):
        """ Images given wavefront assuming fully coherent light. The image
        is obtained as squared absolute value of convolution of input wavefront
         field and ATF of the system. For computational efficiency fourier
        transform convolution theorem is employed.

        :param wf: wavefront to be imaged
        :type wf: pyoptica.Wavefront
        :return: image
        :rtype: pyoptica.Image
        """
        G_g = utils.fft(wf.wavefront)
        G_i = G_g * self.atf
        u_i = utils.ifft(G_i)
        pix_scale = self.pixel_scale * self.m
        image = Image(self.wavelength, pix_scale, self.npix)
        image.image = np.abs(u_i) ** 2
        return image

    def _image_wavefront_fully_incoherent(self, wf):
        """ Images given wavefront assuming fully incoherent light. The image
        is obtained as a convolution of normalized intensity of the input
        wavefront and OTF of the system. For computational efficiency fourier
        transform convolution theorem is employed.

        :param wf: wavefront to be imaged
        :type wf: pyoptica.Wavefront
        :return: image
        :rtype: pyoptica.Image
        """
        I_g = wf.intensity
        I_g /= I_g.max()
        G_g = utils.fft(wf.intensity)
        G_i = G_g * self.otf
        I_i = utils.ifft(G_i)
        pix_scale = self.pixel_scale * self.m
        image = Image(self.wavelength, pix_scale, self.npix)
        image.image = np.real(I_i)
        return image

    def _image_wavefront_partially_coherent(self, wf):
        raise NotImplementedError(
            'Partial coherence simulation has not been implemented yet.')

    def _check_wavefront_compatibility(self, wf):
        """Checks if wavefront is compatible with the system
        :param wf: Wavefront
        :type wf: pyoptica.Wavefront
        :raises: RuntimeError

        """
        if self.npix != wf.npix:
            raise RuntimeError('Imaging system npix != wavefront npix'
                               f'{self.npix} != {wf.npix}')
        if self.wavelength != wf.wavelength:
            raise RuntimeError(
                'Imaging system wavelength != wavefront wavelength'
                f'{self.wavelength} != {wf.wavelength}')
        if self.pixel_scale != wf.pixel_scale:
            raise RuntimeError(
                'Imaging system wavelength != wavefront wavelength'
                f'{self.pixel_scale} != {wf.pixel_scale}')

    def _aberrations_plot_special(self):
        rho, theta = utils.cart2pol(self.f, self.g)
        rho = rho.to(1 / u.m) / (self.na / self.wavelength).to(1 / u.m)
        wavefront = self.wavefront
        wavefront[rho > 1] = np.nan
        return wavefront



import codecs
import os.path

from setuptools import setup

REQUIREMENTS = [
    'astropy', 'matplotlib', 'numpy', 'scipy'
]

PROJECT_URLS = {
    'Blog': 'https://pyoptica.gitlab.io/pyoptica-blog/'
}


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


def readme():
    """
    This function just return the content of README.md
    """
    with open('README.md') as f:
        return f.read()


setup(
    name='pyoptica',
    version=get_version("pyoptica/__init__.py"),
    description='Tools to apply light propagation algorithms',
    long_description_content_type='text/markdown',
    long_description=readme(),
    classifiers=[
        'Programming Language :: Python :: 3.7',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Mathematics'
    ],
    keywords='fourier_optics light_propagation ',
    url='https://gitlab.com/pyoptica',
    project_url=PROJECT_URLS,
    author='Maciej Grochowicz, Michal Miler',
    author_email='pyoptica@protonmail.com',
    license='MIT License',
    packages=['pyoptica', 'pyoptica.optical_elements'],
    install_requires=REQUIREMENTS,
    include_package_data=True,
    package_data={'': ['data/logos/*.png']},
    zip_safe=False,
)



import astropy.units as u

from . import logging
from .optical_elements import Diffuser, FreeSpace

logger = logging.get_standard_logger(__name__)


@u.quantity_input(z=u.m)
def retrieve_phase_gs(
        wf, target, z, max_iter=10, randomize_phase=True,
        return_intermediate=False
):
    """ Iteratively retrieves phase of the wavefront based on the intensity
    distribution in two known planes. Implementation of Gerchberg-Saxton
    phase retrieval algorithm [1]. Phase is retrieved iteratively by
    propagating between input and target planes (limited by `max_iter`)
    and updating the intensity distribution:

    .. image:: ../source/imgs/gs_docstring.png

    :param wf: input wavefront (with initial intensity distribution)
    :type wf: pyoptica.Wavefront
    :param target: desired target distribution (2D array)
    :type target: numpy.array
    :param z: distance between two planes
    :type z: astropy.Quantity of type length
    :param max_iter: number of maximal iterations (default 10)
    :type max_iter: int
    :param randomize_phase: should phase be randomized?  (default True)
    :type randomize_phase: bool
    :param return_intermediate: should intermediate results (after each
        iteration) be returned? (default false)
    :type return_intermediate: bool
    :return: wavefront with retrieved phase (at input plane), list with
        intermediate results (empty if `return_intermediate` is `False`.)
    :rtype: tuple(pyoptica.Wavefront, list(pyoptica.Wavefront))

    **Example**

    >>> import astropy.units as u
    >>>
    >>> import pyoptica as po
    >>> import pyoptica.holography as poh
    >>>
    >>> wavelength = 500 * u.nm
    >>> pixel_scale = 22 * u.um
    >>> npix = 512
    >>> wf = po.Wavefront(wavelength, pixel_scale, npix)

    Let's retrieve phase distribution (or a hologram) that will produce
    at target plane at `z` intenisty distribution given by a binary grating:

    >>> period = 1 * u.mm
    >>> duty = 0.5
    >>> grating = po.BinaryGrating(period, duty)
    >>> target = (wf * grating).intensity

    Now it's time for GS algorithm. We are not going to ask for intermediate
    results:

    >>> z = 10 * u.cm
    >>> holo, _ = poh.retrieve_phase_gs(wf, target, z)
    >>> po.plotting.plot_wavefront(holo, 'phase')
    >>> po.plotting.plot_wavefront(holo * po.FreeSpace(z))

    **References**

    [1] R. W. Gerchberg and W. O. Saxton (1972) -
    "A practical algorithm for the determination of the phase from image
    and diffraction plane pictures,” Optik 35, 237

    """
    source = wf.intensity[:, :]
    wf_work_copy = wf.copy()  # input wavefront should not be manipulated
    if randomize_phase:
        wf_work_copy *= Diffuser()
    intermediate_results = []
    forward_prop_fs = FreeSpace(z)
    backward_prop_fs = FreeSpace(-z)
    for i in range(max_iter):
        wf_work_copy *= forward_prop_fs
        wf_work_copy.intensity = target
        wf_work_copy *= backward_prop_fs
        wf_work_copy.intensity = source
        if return_intermediate:
            intermediate_results.append(wf_work_copy.copy())
        logger.info(f"Done iteration {i + 1} out of {max_iter}.")
    return wf_work_copy, intermediate_results



import astropy.units as u
import numpy as np

from .base_optical_element import BaseOpticalElement
from .. import utils
from ..wavefront import Wavefront


class FreeSpace(BaseOpticalElement):
    """ A class representing an free space for propagation.

    :param distance: Length of the free space to propagate.
    :type distance: astropy.Quantity of type length.
    :param method: Method of prpagation calculation.
    :type method: str

    **Example**

    >>> import astropy.units as u
    >>> import pyoptica as po
    >>>
    >>> fs = po.FreeSpace(10 * u.cm, 'ASPW')

    **References**

    [1] Joseph W. Goodman (2004) -
    *Introduction to Fourier Optics*, W. H. Freeman

    """
    METHODS = ['ASPW']  # , 'Fresnel', 'Fraunhofer']

    def __init__(self, distance, method='ASPW'):

        if method not in self.METHODS:
            raise ValueError(f"Valid methods are: {self.METHODS}")

        self.distance = distance
        self.method = method

    def __mul__(self, wavefront):
        return self.propagate(wavefront)

    __rmul__ = __mul__

    @classmethod
    def calc_optimal_distance(cls, wavefront):
        r"""Based on [2], this function calulates the optimal distance of
        propagation:
        :math:`z = \frac{N\left( \Delta x \right)^2}{\lambda}`
        for given `wavefront`.

        :return: The optimal distance of propagate
        :rtype: astropy.Quantity of type length

        >>> import astropy.units as u
        >>> import pyoptica as po
        >>>
        >>> wavelength = 633 * u.nm
        >>> pixel_scale = 100 * u.nm
        >>> npix = 2048
        >>> distance = 60 * u.um
        >>> wf = po.Wavefront(wavelength, pixel_scale, npix)
        >>> fs = po.FreeSpace(distance, "ASPW")
        >>> fs.calc_optimal_distance(wf)
        <Quantity 32.35387 um>

        """

        pixel_scale = wavefront.pixel_scale
        wavelength = wavefront.wavelength
        npix = wavefront.npix
        z = (pixel_scale ** 2 * npix / wavelength).decompose()
        return z

    @u.quantity_input(distance=u.m)
    def calc_propagation_steps(self, wavefront):
        r"""For the given `wavefront` it calculates number of steps and their size
        required to propagate to the given `distance` based on the sampling
        requirements:

        :math:`z_{sampling} = \frac{N\left( \Delta x \right)^2}{\lambda}`

        :math:`n = \left \lceil  \frac{z}{z_{sampling}} \right \rceil`

        :math:`z_{size} = \frac{z}{n}`

        :param wavefront: wavefront propagating through free space.
        :type wavefront: pyoptica.Wavefront
        :return: number of required steps and their size
        :rtype: Tuple(int, astropy.Qunatity of type length)

        >>> import astropy.units as u
        >>> import pyoptica as po
        >>>
        >>> wavelength = 633 * u.nm
        >>> pixel_scale = 100 * u.nm
        >>> npix = 2048
        >>> distance = 60 * u.um
        >>> wf = po.Wavefront(wavelength, pixel_scale, npix)
        >>> fs = po.FreeSpace(distance)
        >>> fs.calc_propagation_steps(wf)
        (2, <Quantity 30 um>)

        """
        abs_distance = np.abs(self.distance)
        optimal_distance = self.calc_optimal_distance(wavefront)
        no_steps = np.ceil((abs_distance / optimal_distance).decompose()).value
        no_steps = int(no_steps)
        step_size = self.distance / no_steps
        return no_steps, step_size

    def propagate(self, wavefront):
        r"""A method for propagate using Angular Spectrum of Plane Waves. It is
        relatively straightforward to compute; following [1] (Chapter 5.1):

        :math:`U_{2}(x, y)=\mathfrak{F}^{-1}\left\{\mathfrak{F}
        \left\{U_{1}(x, y)\right\} H\left(f_{X}, f_{Y}\right)\right\}`

        where:

        :math:`U_{1}(x, y)` -- the initial wavefront,

        :math:`H\left(f_{X}, f_{Y}\right)=e^{j k z} \exp \left[-j \pi
        \lambda z\left(f_{X}^{2}+f_{Y}^{2}\right)\right]`
        -- the transfer fuction.

        It is computed using the following convolution theorem:

        :math:`f * g =\mathfrak{F}^{-1}\left\{\mathfrak{F}
        \{ f \}\mathfrak{F} \{g\} \right\}`

        Important note: the distance of propagate is limited by sampling [2]:
        :math:`z \leq \frac{N\left( \Delta x \right)^2}{\lambda}`
        If the given `distance` to propagate is greater than the limit given by
        sampling the propagate will be done in multiple equidistant steps
        (summing to the given `distance`).

        :param wavefront: Wavefront to propaget
        :type wavefront: pyoptica.Wavefront
        :return: Propagated wavefront
        :rtype: pyoptica.Wavefront

        **References**

        [1] David Voelz (2011) -
        *Computational Fourier Optics Matlab Tutorial*, Spie Press

        [2] David Voelz and Michael C. Roggemann (2009) -
        *Digital simulation of scalar optical diffraction: revisiting chirp
        function sampling criteria and consequences*, Applied Optics

        >>> import astropy.units as u
        >>> import pyoptica as po
        >>>
        >>> wavelength = 633 * u.nm
        >>> pixel_scale = 100 * u.nm
        >>> npix = 2048
        >>> distance = 60 * u.um
        >>> wf = po.Wavefront(wavelength, pixel_scale, npix)
        >>> fs = po.FreeSpace(distance)
        >>> wf_at_distance = wf * fs

        """
        propagated_wf = Wavefront(wavefront.wavelength, wavefront.pixel_scale, wavefront.npix)
        steps_number, step_size = self.calc_propagation_steps(wavefront)
        h = FreeSpace(step_size, method=self.method).phase_transmittance(wavefront)
        H = utils.fft(h)
        wf_at_distance = utils.fft(wavefront.wavefront)
        for _ in range(steps_number):
            wf_at_distance = wf_at_distance * H
        wf_at_distance = utils.ifft(wf_at_distance)
        propagated_wf.wavefront = wf_at_distance
        self.logger.info(
            f"Propagated using '{self.method}' to {self.distance:.0}.")
        return propagated_wf

    def amplitude_transmittance(self, wavefront):
        """We assumed that the free space amplitude transmittance is one .

        :param wavefront: The wavefront which interacts with the free space.
        :type wavefront: pyoptica.Wavefront
        :return: distribution of amplitude transmittance resulting from the free space.
        :rtype: numpy.array

        """

        return np.ones_like(wavefront.amplitude)

    def phase_transmittance(self, wavefront):
        """We assumed that the perfect diffuser introduces completely random phase.

        :param wavefront: The wavefront which interacts with the free space.
        :type wavefront: pyoptica.Wavefront
        :return: distribution of phase transmittance resulting from the free space.
        :rtype: numpy.array
        """
        H = np.zeros_like(wavefront.wavefront)
        pix_scale_m = wavefront.pixel_scale.to(u.m).value
        wavelen_m = wavefront.wavelength.to(u.m).value
        distance_m = self.distance.to(u.m).value

        x = wavefront.x.to(u.m).value
        y = wavefront.y.to(u.m).value
        npix = wavefront.npix
        f_x = (x / (pix_scale_m ** 2 * npix))
        f_y = (y / (pix_scale_m ** 2 * npix))
        k = wavefront.k.to(1 / u.m).value

        if self.method == "ASPW":
            rhosqr = f_x ** 2 + f_y ** 2
            exp_1 = 1.j * k * distance_m
            exp_2 = -1.j * np.pi * wavelen_m * distance_m * rhosqr
            H = np.exp(exp_1 + exp_2)
        return utils.ifft(H)
